import os
import time
import logging

import multiprocessing

import pickle
import gzip
import bz2

from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

import numpy as np
import pandas as pd

from pathlib import Path

from sklearn.model_selection import train_test_split

def run_processor(fun, *args, **kwargs):
    """Run a function in processor using multiprocessing"""
    # Input: function to be run, args to the
    # Return: Output of the input fun
    def run_fun(fun, return_dict, args, kwargs):
        return_dict[1] = fun(*args, **kwargs)
    manager = multiprocessing.Manager()
    return_dict = manager.dict()
    b = multiprocessing.Process(target=run_fun, args=(fun, return_dict, args, kwargs))
    b.start()
    b.join()
    return return_dict[1]


def mat2onehot(mat, category = None):
    """ Turns matrix into one hot format"""
    mat = np.array(mat)
    if category == None:
        category = np.unique(mat)
    mat_onehot = [np.isin(category, array) for array in mat]
    mat_onehot = np.array(mat_onehot).astype('int')
    return mat_onehot


def multiple_hot_df(item_lists):
    """ Merge data frames """
    fullset = []
    for item_list in item_lists:
        fullset += np.array(item_list)[np.invert(np.isin(item_list, fullset))].tolist()
    mat_bin = np.array([np.isin(fullset,item_list) for item_list in item_lists]).astype('int')
    mat_bin = pd.DataFrame(mat_bin.transpose())
    mat_bin.index = fullset
    return mat_bin


def makedirs(logdir):
    Path(logdir).mkdir(parents = True, exist_ok = True)


def GetSpacedElements(array, numElems = 4):
    out = array[np.round(np.linspace(0, len(array)-1, numElems)).astype(int)]
    return out


def save_pickle(obj, filename, zip_method = None, protocol=-1, correct_suffix = True):
    """"""
    if correct_suffix:
        filename = filename.rsplit(".",1)[0]
        if zip_method is None:
            filename += ".pickle"
        elif zip_method is 'gzip':
            filename += ".pgz"
        elif zip_method is 'bz2':
            filename += ".pbz2"
        else:
            raise Exception('Specify a correct zip_method: {}'.format(zip_method))

    if zip_method is None:
        with open(filename, 'wb') as f:
            pickle.dump(obj, f, protocol)
    elif zip_method == 'gzip':
        with gzip.open(filename, 'w') as f:
            pickle.dump(obj, f)
    elif zip_method == 'bz2':
         with bz2.BZ2File(filename, 'w') as f:
            pickle.dump(obj, f)
    else:
        raise Exception('Specify a correct zip_method: {}'.format(zip_method))


def load_pickle(filename, zip_method=None, correct_suffix = True):

    if correct_suffix:
        filename = filename.rsplit(".",1)[0]
        if zip_method is None:
            filename += ".pickle"
        elif zip_method is 'gzip':
            filename += ".pgz"
        elif zip_method is 'bz2':
            filename += ".pbz2"
        else:
            raise Exception('Specify a correct zip_method: {}'.format(zip_method))

    if zip_method is None:
        with open(filename, 'rb') as f:
            loaded = pickle.load(f)
    elif zip_method == 'gzip':
        with gzip.open(filename, 'r') as f:
            loaded = pickle.load(f)
    elif zip_method == 'bz2':
         with bz2.BZ2File(filename, 'r') as f:
            loaded = pickle.load(f)
    else:
        raise Exception('Specify a correct zip_method: {}'.format(zip_method))

    return loaded


def reduce_dimensions(X, reduced_dimension = 2, method = 'PCA', tsne_min = 50, match_dims = True):
    """
    Reduce dimensioanlity of X using dimensionality reduction method.
    """
    if X.shape[-1] > reduced_dimension:

        if method == "PCA":
            pca = PCA(n_components = reduced_dimension)
            pca.fit(X)
            X_pca = pca.transform(X)
            X_reduced = X_pca[:,:reduced_dimension]
            dim_name = 'pc'

        elif method == "tSNE":
            if X.shape[-1] > tsne_min:
                X,_ = reduce_dimensions(X, reduced_dimension = 50, method = 'PCA')
            tsne = TSNE(n_components = reduced_dimension, init = 'pca')
            X_tsne = tsne.fit_transform(X)
            X_reduced = X_tsne[:,:2]
            dim_name = 't-SNE'
    else:

        if X.shape[-1] < reduced_dimension and match_dims:
            dim_diff = reduced_dimension - X.shape[-1]
            X_reduced = np.concatenate([X, np.zeros([len(X),dim_diff])], axis = 1)
        else:
            X_reduced = X

        dim_name = "dim"

    dim_labels = ["{}_{}".format(dim_name, ii) for ii in range(reduced_dimension)]
    return X_reduced, dim_labels


def save_df_as_npz(filename, df):
    np.savez(filename, matrix = df.values, columns = df.columns, index = df.index, allow_pickle = True)


def load_df_from_npz(filename):
    uploaded = np.load(filename, allow_pickle = True)
    df = pd.DataFrame(uploaded['matrix'], columns = uploaded['columns'], index = uploaded['index'])
    return df


def reduce_samples(X, label_in, num_reduced = 10000, type='sample'):
    """
    type = ['sample', 'PCA']
    """

    if X.shape[1] > num_reduced:

        if type == 'sample':
            test_size= num_reduced/len(X)
            if test_size < 1:
                _,X_new,_,_ = train_test_split(X,label_in,test_size=test_size,stratify=label_in)

        elif type == 'PCA':
            X_t         = X.transpose()
            num_reduced = min(num_reduced, X_t.shape[0])
            X_new , _   = reduce_dimensions(X_t, reduced_dimension = num_reduced, method = 'PCA')
            X_new       = X_new.transpose()

        else:
            raise Exception('Input valid VAE_feature_input ')

    else:

        X_new = X

    return X_new



import os

import math
import numpy as np

import pandas as pd

from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import scale
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA

import pickle
import re

import copy

def split_index(length, k_split, label = None, random_seed = 0):
    """ Given data and label, split into test/train then give result """
    list_index = []
    X0 = np.arange(length)
    if k_split < 1:
        if label is None:
            label = np.ones(length)
        test_data, train_data, _, _ = train_test_split(pd.DataFrame(X0), label,
                                                       test_size    = k_split,
                                                       stratify     = label,
                                                       random_state = random_seed)
        split_values = (np.array(train_data.index), np.array(test_data.index))
        list_index.append(split_values)
    #
    elif k_split == 1:
        split_values = (X0, X0)
        list_index.append(split_values)
    #
    else:
        skf = StratifiedKFold(n_splits     = int(k_split),
                              random_state = random_seed,
                              shuffle      = True)
        for split_values in skf.split(X0, label):
            list_index.append(split_values)
    #
    return list_index


def split_data(data_list, split_idx, axis_list = None):
    """ Split list of data by train/test index for specificed axis"""
    train_index, test_index = split_idx
    dataset = []
    #
    if axis_list is None:
        axis_list = [0] * len(data_list)
    #
    for data in data_list:
        dataset.append(data[train_index])
        dataset.append(data[test_index])
    #
    return dataset


def kfold_split(X,y, k_split, random_seed = 1, isPrecompMatrix = False):
    skf = StratifiedKFold(n_splits     = k_split,
                          random_state = random_seed,
                          shuffle      = True)
    if not isPrecompMatrix:
        datasets = [[X[idx_train],
                     X[idx_test],
                     y[idx_train],
                     y[idx_test]] for idx_train, idx_test in skf.split(X, y)]
    else:
        datasets = [[X[idx_train][:,idx_train],
                     X[idx_test][:,idx_train],
                     y[idx_train],
                     y[idx_test]] for idx_train, idx_test in skf.split(X, y)]
    return datasets


class data_handler(object):
    """
    Main Data (X, y)
    """
    def __init__(self, h_dims = None, X = None, y = None, label = None):
        self.X = X
        self.y = y
        self.iterator = None
    #
    def create_split_index_list(self, k_split, random_seed = 0):
        """ """
        self.split_index_list = split_index(length      = len(self.X),
                                            k_split     = k_split,
                                            label       = self.X.obs['Labels'],
                                            random_seed = random_seed)
    #
    def copy(self):
        return(copy.deepcopy(self))
    #
    def create_dataset(self, kfold_idx=None):
        """
        split_idx: index of the training and test set in X and y data
        """
        if kfold_idx is None:
            split_idx = self.split_idx
        else:
            split_idx = self.get_kfold_index(kfold_idx)
            self.split_idx = split_idx
        data_list = [self.X.X,
                     self.y.X]
        self.dataset = split_data(data_list, split_idx)
    #
    def get_kfold_index(self, kfold_idx):
        return self.split_index_list[kfold_idx]



"""
A python logger

Usage:
    # Set logger verbose level.
    import os
    os.environ['VERBOSE'] = 1

    import logger
    log = logger.get('../logs/sample_log')
    
    log.info('Hello world!')
    log.info('Hello again!', verbose=2)
    log.warning('Something might be wrong.')
    log.error('Something is wrong.')
    log.fatal('Failed.')
"""

from __future__ import print_function
import datetime
import inspect
import os
import sys
import threading
import traceback

TERM_COLOR = {
    'normal': '\033[0m',
    'bright': '\033[1m',
    'invert': '\033[7m',
    'black': '\033[30m',
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'magenta': '\033[35m',
    'cyan': '\033[36m',
    'white': '\033[37m',
    'default': '\033[39m'
}

log = None
log_lock = threading.Lock()


def get(fname=None):
  """
  Returns a logger instance, with optional log file output.
  """
  global log
  if log is not None and fname is None:
    return log

    # fname = os.environ.get('LOGTO', None)
    # if fname is None:
    #     fname = default_fname
  else:
    log = Logger(fname)

    return log


class Logger(object):

  def __init__(self, filename=None, default_verbose=0):
    """
    Constructs a logger with optional log file output.

    Args:
        filename: optional log file output. If None, nothing will be 
        written to file
    """
    now = datetime.datetime.now()
    self.verbose_thresh = int(os.environ.get('VERBOSE', 0))
    self.default_verbose = default_verbose
    if filename is not None:
      self.filename = filename
      dirname = os.path.dirname(self.filename)
      if not os.path.exists(dirname):
        os.makedirs(dirname)
      open(self.filename, 'w').close()
      self.info('Log written to {}'.format(os.path.abspath(self.filename)))
    else:
      self.filename = None

    pass

  @staticmethod
  def get_time_str(t=None):
    """
    Returns a formatted time string.

    Args:
        t: datetime, default now.
    """
    if t is None:
      t = datetime.datetime.now()

    # timestr = '{:04d}{:02d}{:02d}-{:02d}{:02d}{:02d}'.format(
    #     t.year, t.month, t.day, t.hour, t.minute, t.second)[2:]
    timestr = t.isoformat(' ')
    # timestr = '{:04d}{:02d}{:02d}{:02d}{:02d}{:02d}'.format(
    #     t.year, t.month, t.day, t.hour, t.minute, t.second)
    # timestr = '{:02d}:{:02d}:{:02d}'.format(t.hour, t.minute, t.second)

    return timestr

  def log(self, message, typ='info', verbose=None):
    """
    Writes a message.

    Args:
        message: string, message content.
        typ: string, type of the message. info, warning, error, or fatal.
        verbose: number, verbose level of the message. If lower than the 
        environment variable, then the message will be logged to standard 
        output and log output file (if set).
    """
    threadstr = '{}'.format(threading.current_thread().ident)[-4:]
    if typ == 'info':
      typstr_print = '{}I{}{}'.format(TERM_COLOR['green'], threadstr,
                                      TERM_COLOR['default'])
      typstr_log = 'I{}'.format(threadstr)
    elif typ == 'warning':
      typstr_print = '{}W{}{}'.format(TERM_COLOR['yellow'], threadstr,
                                      TERM_COLOR['default'])
      typstr_log = 'W{}'.format(threadstr)
    elif typ == 'debug':
      typstr_print = '{}D{}{}'.format(TERM_COLOR['yellow'], threadstr,
                                      TERM_COLOR['default'])
      typstr_log = 'D{}'.format(threadstr)
    elif typ == 'error':
      typstr_print = '{}E{}{}'.format(TERM_COLOR['red'], threadstr,
                                      TERM_COLOR['default'])
      typstr_log = 'E{}'.format(threadstr)
    elif typ == 'fatal':
      typstr_print = '{}F{}{}'.format(TERM_COLOR['red'], threadstr,
                                      TERM_COLOR['default'])
      typstr_log = 'F{}'.format(threadstr)
    else:
      raise Exception('Unknown log type: {0}'.format(typ))
    timestr = self.get_time_str()
    for (frame, filename, line_number, function_name, lines, index) in \
            inspect.getouterframes(inspect.currentframe()):
      fn = os.path.basename(filename)
      if fn != 'logger.py':
        break
    cwd = os.getcwd()
    if filename.startswith(cwd):
      filename = filename[len(cwd):]
    filename = filename.lstrip('/')

    callerstr = '{}:{}'.format(filename, line_number)
    if len(callerstr) > 20:
      callerstr = '...{}'.format(callerstr[-17:])
    printstr = '{} {} {} {}'.format(typstr_print, timestr, callerstr, message)
    logstr = '{} {} {} {}'.format(typstr_log, timestr, callerstr, message)

    if verbose is None:
      verbose = self.default_verbose

    if type(verbose) != int:
      raise Exception('Unknown verbose value: {}'.format(verbose))
    # print((self.verbose_thresh, type(self.verbose_thresh), verbose, type(verbose)))
    # print(threadstr, 'Waiting log lock')
    log_lock.acquire()
    try:
      if self.verbose_thresh >= verbose:
        print(printstr)

      if self.filename is not None:
        with open(self.filename, 'a') as f:
          f.write(logstr)
          f.write('\n')
    except e:
      print('Error occurred!!')
      print(str(e))
    finally:
      log_lock.release()
    # print(threadstr, 'Released log lock')
    pass

  def info(self, message, verbose=None):
    """
    Writes an info message.

    Args:
        message: string, message content.
        verbose: number, verbose level.
    """
    self.log(message, typ='info', verbose=verbose)
    pass

  def warning(self, message, verbose=1):
    """
    Writes a warning message.

    Args:
        message: string, message content.
        verbose: number, verbose level.
    """
    self.log(message, typ='warning', verbose=verbose)
    pass

  def error(self, message, verbose=0):
    """
    Writes an info message.

    Args:
        message: string, message content.
        verbose: number, verbose level.
    """
    self.log(message, typ='error', verbose=verbose)
    pass

  def debug(self, message, verbose=None):
    self.log(message, typ='debug', verbose=verbose)
    pass

  def fatal(self, message, verbose=0):
    """
    Writes a fatal message, and exits the program.

    Args:
        message: string, message content.
        verbose: number, verbose level.
    """
    self.log(message, typ='fatal', verbose=verbose)
    sys.exit(0)
    pass

  def log_args(self, verbose=None):
    self.info('Command: {}'.format(' '.join(sys.argv)))
    pass

  def log_exception(self, exception):
    tb_str = traceback.format_exc(exception)
    self.error(tb_str)
    pass

  def verbose_level(self, level):

    class VerboseScope():

      def __init__(self, logger, new_level):
        self._new_level = new_level
        self._logger = logger
        pass

      def __enter__(self):
        self._restore = self._logger.default_verbose
        self._logger.default_verbose = self._new_level
        pass

      def __exit__(self, type, value, traceback):
        self._logger.default_verbose = self._restore
        pass

    return VerboseScope(self, level)



## Template for saving

import tensorflow as tf

def array_to_tfrecords(X, y, output_file):
    feature = {
    'X': tf.train.Feature(float_list=tf.train.FloatList(value=X.flatten())),
    'y': tf.train.Feature(float_list=tf.train.FloatList(value=y.flatten()))
    }
    example = tf.train.Example(features=tf.train.Features(feature=feature))
    serialized = example.SerializeToString()
    writer = tf.python_io.TFRecordWriter(output_file)
    writer.write(serialized)
    writer.close()


def parse_proto(example_proto):
    features = {
    'X': tf.FixedLenFeature([3], tf.float32)
    }
    parsed_features = tf.parse_single_example(example_proto, features)
    return([parsed_features['X']])


def parse_proto(example_proto):
    features = {
    'X': tf.FixedLenFeature((345,), tf.float32),
    'y': tf.FixedLenFeature((5,), tf.float32),
    }
    parsed_features = tf.parse_single_example(example_proto, features)
    return(parsed_features['X'], parsed_features['y'])


def read_tfrecords(file_names=("file1.tfrecord", "file2.tfrecord", "file3.tfrecord"),
                   buffer_size=10000,
                   batch_size=100):
    dataset = tf.contrib.data.TFRecordDataset(file_names)
    dataset = dataset.map(parse_proto)
    dataset = dataset.shuffle(buffer_size)
    dataset = dataset.repeat()
    dataset = dataset.batch(batch_size)
    return(tf.contrib.data.Iterator.from_structure(dataset.output_types, dataset.output_shapes))



import os
import time
import logging

import tensorflow as tf

from . import VAE
from . import siVAE
from . import siVAE2

from .analysis import run_FI

from siVAE import FeatureImportance as FI

import time
import numpy as np

def run_VAE(graph_args_sample, LE_method, datah_sample, datah_feature = None, do_pretrain = True,
            graph_args_feature = None, graph_args_siVAE = None, rs = 0, zv_recon_scale = 0,
            method_DE = ['SaliencyMaps','IntGrad','GradInput', 'Shapley', 'DeepLIFT'],
            sample_set = None, do_FA = False, kwargs_FI = {}, siVAE_version = 1):
    """
    run either VAE or siVAE and return the results in results_dict

    Parameters
    ----------
    graph_args_sample
        Arguments for sample-wise encoder/decoder
    graph_args_feature
        Arguments for feature-wise encoder/decoder
    graph_args_siVAE
        Arguments for combined model
    LE_method
        Model to use, either VAE or siVAE
    datah_sample
        data_handler object for sample-wise encoder/decoder
    datah_feature
        data_handler object for feature-wise encoder/decoder
    do_pretrain
        Set True to pre-train sample-wise and feature-wise encoder/decoder for siVAE
    method_DE
        List of feature attribution methods to use
        ['SaliencyMaps','IntGrad','GradInput', 'Shapley', 'DeepLIFT']
    sample_set
        Sample set to use for feature attribution
    do_FA
        Set True to run feature attribution
    kwargs_FI
        Keyword arguments for feature attribution
    """

    from . import siVAE
    from . import siVAE2

    if siVAE_version == 2:
        siVAE = siVAE2

    method_DE_dict = {"SaliencyMaps": "Saliency Maps",
                      "GradInput"   : "Gradient * Input",
                      "IntGrad"     : "Integrated Gradients",
                      'Epsilon-LRP' : 'Epsilon-LRP',
                      'DeepLIFT'    : 'DeepLIFT',
                      'Occlusion'   : 'Occlusion',
                      'Shapley'     : 'Shapley Value'}

    method_DE = [ method_DE_dict[method] for method in method_DE ]

    if sample_set is not None:
        sample_input = sample_set['samples']
    else:
        sample_input = None

    ## Set up architecture of hidden layers
    LE_dim       = graph_args_sample['LE_dim']
    architecture = graph_args_sample['architecture']

    # For datah_sample
    h_dims = architecture.replace('LE',str(LE_dim))
    h_dims = [int(dim) for dim in h_dims.split("-")]
    datah_sample.index_latent_embedding = int(h_dims.pop())
    datah_sample.h_dims = h_dims

    # For datah_feature
    if datah_feature is not None:
        h_dims = architecture.replace('LE',str(LE_dim))
        h_dims = [int(dim) for dim in h_dims.split("-")]
        datah_feature.index_latent_embedding = int(h_dims.pop())
        _ = h_dims.pop(-2)
        datah_feature.h_dims = h_dims

    ## Run Model
    if LE_method == 'siVAE':

        # Graph argument for feature-wise encoder/decoder
        if graph_args_feature is None:
            graph_args_feature = dict(graph_args_sample)
            graph_args_feature['var_dependency'] = True
            graph_args_feature['X_mu_use_bias'] = True

        # Graph argument for combined model
        if graph_args_siVAE is None:
            graph_args_siVAE = dict(graph_args_sample)
            graph_args_siVAE['zv_recon_scale'] = zv_recon_scale
            graph_args_siVAE['early_stopping'] = 100

        print('siVAE?')
        result, result_s, result_f, model = siVAE.run(datah_sample, datah_feature,
                                                      graph_args_sample,
                                                      graph_args_feature,
                                                      graph_args_siVAE,
                                                      rs = rs,
                                                      do_pretrain  = do_pretrain,
                                                      sample_input = sample_input,
                                                      do_FA        = do_FA,
                                                      method_DE    = method_DE,
                                                      kwargs_FI    = kwargs_FI)

        result_dict = {'sample':result_s,'feature':result_f,'model':result}

    else:

        if datah_sample.iterator is None:
            tf.reset_default_graph()

        config = graph_args_sample['config']

        with tf.Session(config = config) as sess:

            with FI.DeepExplain(session=sess) as de:

                model = VAE.AutoEncoder(data_handler = datah_sample, random_seed = rs, isVAE = True, name = 'VAE', **graph_args_sample)
                model.build_model(reset_graph = False)
                result = model.train(sess)

                ## Additional analysis
                if sample_input is not None:
                    sample_dict = {}
                    if len(method_DE) > 0 and do_FA:
                        logging.info('Perform feature attribution')
                        time.sleep(1)

                        feed_dict_full = {model.X  : sample_input,
                                          model.X_target : sample_input}

                        sample_z = model.calculate_z_mu(feed_dict_full)

                        feed_dict_base = {}
                        attributions_dict_all = {}

                        ## Encoder
                        attributions_dict = run_FI(input     = model.X,
                                                   target    = model.z_mu,
                                                   sample_DE = sample_input,
                                                   de        = de,
                                                   sess      = sess,
                                                   method_DE = method_DE,
                                                   feed_dict = feed_dict_base,
                                                   kwargs_FI = kwargs_FI,
                                                   mode      = 'reverse')

                        attributions_dict_all['encoder'] = attributions_dict

                        # ## Decoder
                        kwargs_FI2 = dict(kwargs_FI)
                        _ = kwargs_FI2.pop('baseline',None)
                        attributions_dict = run_FI(input     = model.z_sample,
                                                   target    = model.X_dist.mu,
                                                   sample_DE = sample_z,
                                                   de        = de,
                                                   sess      = sess,
                                                   method_DE = method_DE,
                                                   feed_dict = feed_dict_base,
                                                   kwargs_FI = kwargs_FI2,
                                                   mode      = 'forward')

                        attributions_dict_all['decoder'] = attributions_dict

                        sample_dict['attributions_samples'] = attributions_dict_all

                    result['sample_dict'] = sample_dict

        result_dict = {'model':result}

    result_dict['model']['metadata'] = np.array(datah_sample.X.obs_names)

    return result_dict



"""
Implements second-order matrix vector product using R-op (forward-mode
automatic differentiation) in TensorFlow.

1) Hessian
2) Fisher
3) Gauss-Newton
"""
import tensorflow as tf

from tensorflow_forward_ad import logger
from tensorflow_forward_ad.fwgrad import forward_gradients

log = logger.get()


def hessian_vec_fw(ys, xs, vs, grads=None):
  """Implements Hessian vector product using forward on backward AD.

  Args:
    ys: Loss function.
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    Hv: Hessian vector product, same size, same shape as xs.
  """
  # Validate the input
  if type(xs) == list:
    if len(vs) != len(xs):
      raise ValueError("xs and vs must have the same length.")

  if grads is None:
    grads = tf.gradients(ys, xs, gate_gradients=True)
  return forward_gradients(grads, xs, vs, gate_gradients=True)


def hessian_vec_bk(ys, xs, vs, grads=None):
  """Implements Hessian vector product using backward on backward AD.

  Args:
    ys: Loss function.
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    Hv: Hessian vector product, same size, same shape as xs.
  """
  # Validate the input
  if type(xs) == list:
    if len(vs) != len(xs):
      raise ValueError("xs and vs must have the same length.")

  if grads is None:
    grads = tf.gradients(ys, xs, gate_gradients=True)
  return tf.gradients(grads, xs, vs, gate_gradients=True)


def fisher_vec_bk(ys, xs, vs):
  """Implements Fisher vector product using backward AD.

  Args:
    ys: Loss function, scalar.
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    J'Jv: Fisher vector product.
  """
  # Validate the input
  if type(xs) == list:
    if len(vs) != len(xs):
      raise ValueError("xs and vs must have the same length.")

  grads = tf.gradients(ys, xs, gate_gradients=True)
  gradsv = list(map(lambda x: tf.reduce_sum(x[0] * x[1]), zip(grads, vs)))
  jv = tf.add_n(gradsv)
  jjv = list(map(lambda x: x * jv, grads))
  return jjv


def fisher_vec_fw(ys, xs, vs):
  """Implements Fisher vector product using backward and forward AD.

  Args:
    ys: Loss function or output variables.
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    J'Jv: Fisher vector product.
  """
  # Validate the input
  if type(xs) == list:
    if len(vs) != len(xs):
      raise ValueError("xs and vs must have the same length.")

  jv = forward_gradients(ys, xs, vs, gate_gradients=True)
  jjv = tf.gradients(ys, xs, jv, gate_gradients=True)
  return jjv


def gauss_newton_vec(ys, zs, xs, vs):
  """Implements Gauss-Newton vector product.

  Args:
    ys: Loss function.
    zs: Before output layer (input to softmax).
    xs: Weights, list of tensors.
    vs: List of perturbation vector for each weight tensor.

  Returns:
    J'HJv: Guass-Newton vector product.
  """
  # Validate the input
  if type(xs) == list:
    if len(vs) != len(xs):
      raise ValueError("xs and vs must have the same length.")

  grads_z = tf.gradients(ys, zs, gate_gradients=True)
  hjv = forward_gradients(grads_z, xs, vs, gate_gradients=True)
  jhjv = tf.gradients(zs, xs, hjv, gate_gradients=True)
  return jhjv, hjv


def fisher_vec_z(ys, xs, vs):
  """Implements JJ'v, where v is on the output space.

  Args:
    ys: Loss function or output variables.
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    JJ'v: Fisher vector product on the output space.
  """
  # Validate the input
  if type(ys) == list:
    if len(vs) != len(ys):
      raise ValueError("ys and vs must have the same length.")

  jv = tf.gradients(ys, xs, vs, gate_gradients=True)
  jjv = forward_gradients(ys, xs, jv, gate_gradients=True)
  return jjv


def gauss_newton_vec_z(ys, zs, xs, vs):
  """Implements HJJ'v, where v is on the output space.

  Args:
    ys: Loss function or output variables.
    zs: Before output layer (input to softmax).
    xs: Weights, list of tensors.
    vs: List of tensors to multiply, for each weight tensor.

  Returns:
    HJJ'v: Gauss-Newton vector product on the output space.
  """
  # Validate the input
  if type(zs) == list:
    if len(vs) != len(zs):
      raise ValueError("zs and vs must have the same length.")

  grads_z = tf.gradients(ys, zs, gate_gradients=True)
  jv = tf.gradients(zs, xs, vs, gate_gradients=True)
  hjjv = forward_gradients(grads_z, xs, jv, gate_gradients=True)
  return hjjv



#!/usr/bin/env python

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

import numpy

setup(
    cmdclass={'build_ext': build_ext},
    ext_modules=[
        Extension(
            "cbfs", sources=["cbfs.pyx"], include_dirs=[numpy.get_include()])
    ],)



from unittest import TestCase
import pkg_resources
import logging, warnings
import tensorflow as tf
import numpy as np

from deepexplain.tensorflow import DeepExplain
from deepexplain.tensorflow.methods import original_grad # test only

activations = {'Relu': tf.nn.relu,
                'Sigmoid': tf.nn.sigmoid,
                'Softplus': tf.nn.softplus,
                'Tanh': tf.nn.tanh}


def simple_model(activation, session):
    X = tf.placeholder("float", [None, 2])
    w1 = tf.Variable(initial_value=[[1.0, -1.0], [-1.0, 1.0]])
    b1 = tf.Variable(initial_value=[1.5, -1.0])
    w2 = tf.Variable(initial_value=[[1.1, 1.4], [-0.5, 1.0]])
    b2 = tf.Variable(initial_value=[0.0, 2.0])

    layer1 = activation(tf.matmul(X, w1) + b1)
    out = tf.matmul(layer1, w2) + b2
    session.run(tf.global_variables_initializer())
    return X, out


def simpler_model(session):
    """
    Implements ReLU( ReLU(x1 - 1) - ReLU(x2) )
    :
    """
    X = tf.placeholder("float", [None, 2])
    w1 = tf.Variable(initial_value=[[1.0, 0.0], [0.0, 1.0]], trainable=False)
    b1 = tf.Variable(initial_value=[-1.0, 0], trainable=False)
    w2 = tf.Variable(initial_value=[[1.0], [-1.0]], trainable=False)

    l1 = tf.nn.relu(tf.matmul(X, w1) + b1)
    out = tf.nn.relu(tf.matmul(l1, w2))
    session.run(tf.global_variables_initializer())
    return X, out


def min_model(session):
    """
    Implements min(xi)
    """
    X = tf.placeholder("float", [None, 2])
    out = tf.reduce_min(X,1)
    session.run(tf.global_variables_initializer())
    return X, out


def min_model_2d(session):
    """
    Implements min(xi)
    """
    X = tf.placeholder("float", [None, 2, 2])
    out = tf.reduce_min(tf.reshape(X, (-1, 4)), 1)
    session.run(tf.global_variables_initializer())
    return X, out


def simple_multi_inputs_model(session):
    """
    Implements Relu (3*x1|2*x2) | is a concat op
    :
    """
    X1 = tf.placeholder("float", [None, 2])
    X2 = tf.placeholder("float", [None, 2])
    w1 = tf.Variable(initial_value=[[3.0, 0.0], [0.0, 3.0]], trainable=False)
    w2 = tf.Variable(initial_value=[[2.0, 0.0], [0.0, 2.0]], trainable=False)

    out = tf.nn.relu(tf.concat([X1*w1, X2*w2], 1))
    session.run(tf.global_variables_initializer())
    return X1, X2, out


def simple_multi_inputs_model2(session):
    """
    Implements Relu (3*x1|2*x2) | is a concat op
    :
    """
    X1 = tf.placeholder("float", [None, 2])
    X2 = tf.placeholder("float", [None, 1])
    w1 = tf.Variable(initial_value=[3.0], trainable=False)
    w2 = tf.Variable(initial_value=[2.0], trainable=False)

    out = tf.nn.relu(tf.concat([X1*w1, X2*w2], 1))
    session.run(tf.global_variables_initializer())
    return X1, X2, out


def train_xor(session):
    # Since setting seed is not always working on TF, initial weights values are hardcoded for reproducibility
    X = tf.placeholder("float", [None, 2])
    Y = tf.placeholder("float", [None, 1])
    w1 = tf.Variable(initial_value=[[0.10711301, -0.0987727], [-1.57625198, 1.34942603]])
    b1 = tf.Variable(initial_value=[-0.30955192, -0.14483099])
    w2 = tf.Variable(initial_value=[[0.69259691], [-0.16255915]])
    b2 = tf.Variable(initial_value=[1.53952825])

    l1 = tf.nn.relu(tf.matmul(X, w1) + b1)
    out = tf.matmul(l1, w2) + b2
    session.run(tf.global_variables_initializer())

    # Define loss and optimizer
    loss = tf.reduce_mean(tf.losses.mean_squared_error(Y, out))
    train_step = tf.train.GradientDescentOptimizer(0.05).minimize(loss)

    # Generate dataset random
    np.random.seed(10)
    x = np.random.randint(0, 2, size=(10, 2))
    y = np.expand_dims(np.logical_or(x[:, 0], x[:, 1]), -1)
    l = None
    for _ in range(100):
        l, _, = session.run([loss, train_step], feed_dict={X: x, Y: y})

    return np.abs(l - 0.1) < 0.01


class TestDeepExplainGeneralTF(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_tf_available(self):
        try:
            pkg_resources.require('tensorflow>=1.0.0')
        except Exception:
            try:
                pkg_resources.require('tensorflow-gpu>=1.0.0')
            except Exception:
                self.fail("Tensorflow requirement not met")

    def test_simple_model(self):
        X, out = simple_model(tf.nn.relu, self.session)
        xi = np.array([[1, 0]])
        r = self.session.run(out, {X: xi})
        self.assertEqual(r.shape, xi.shape)
        np.testing.assert_equal(r[0], [2.75,  5.5])

    def test_simpler_model(self):
        X, out = simpler_model(self.session)
        xi = np.array([[3.0, 1.0]])
        r = self.session.run(out, {X: xi})
        self.assertEqual(r.shape, (xi.shape[0], 1))
        np.testing.assert_equal(r[0], [1.0])

    def test_training(self):
        session = tf.Session()
        r = train_xor(session)
        self.assertTrue(r)

    def test_context(self):
        """
        DeepExplain overrides nonlinearity gradient
        """
        # No override
        from deepexplain.tensorflow import DeepExplain

        X = tf.placeholder("float", [None, 1])
        for name in activations:
            x1 = activations[name](X)
            x1_g = tf.gradients(x1, X)[0]
            self.assertEqual(x1_g.op.type, '%sGrad' % name)

        # Override (note: that need to pass graph! Multiple thread testing??)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            for name in activations:
                # Gradients of nonlinear ops are overriden
                x2 = activations[name](X)
                self.assertEqual(x2.op.get_attr('_gradient_op_type').decode('utf-8'), 'DeepExplainGrad')

    def test_mismatch_input_lens(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1 = tf.placeholder("float", [None, 1])
            X2 = tf.placeholder("float", [None, 1])
            w1 = tf.Variable(initial_value=[[0.10711301]])
            w2 = tf.Variable(initial_value=[[0.69259691]])
            out = tf.matmul(X1, w1) + tf.matmul(X2, w2)

            self.session.run(tf.global_variables_initializer())
            with self.assertRaises(RuntimeError) as cm:
                de.explain('grad*input', out, [X1, X2], [[1], [2], [3]])
            self.assertIn(
                'List of input tensors and input data have different lengths',
                str(cm.exception)
            )

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1 = tf.placeholder("float", [None, 1])
            X2 = tf.placeholder("float", [None, 1])
            w1 = tf.Variable(initial_value=[[10.0]])
            w2 = tf.Variable(initial_value=[[10.0]])
            out = tf.matmul(X1, w1) + tf.matmul(X2, w2)

            self.session.run(tf.global_variables_initializer())
            attributions = de.explain('grad*input', out, [X1, X2], [[[2]], [[3]]])
            self.assertEqual(len(attributions), 2)
            self.assertEqual(attributions[0][0], 20.0)
            self.assertEqual(attributions[1][0], 30.0)

    def test_supported_activations(self):
        X = tf.placeholder("float", [None, 3])
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            xi = [[-1, 0, 1]]
            Y = tf.nn.relu(X)
            r = self.session.run(Y, {X: xi})
            np.testing.assert_almost_equal(r[0], [0, 0, 1], 7)
            Y = tf.nn.elu(X)
            r = self.session.run(Y, {X: xi})
            np.testing.assert_almost_equal(r[0], [-0.632120558, 0, 1], 7)
            Y = tf.nn.sigmoid(X)
            r = self.session.run(Y, {X: xi})
            np.testing.assert_almost_equal(r[0], [0.268941421, 0.5, 0.731058578], 7)
            Y = tf.nn.tanh(X)
            r = self.session.run(Y, {X: xi})
            np.testing.assert_almost_equal(r[0], [-0.761594155, 0, 0.761594155], 7)
            Y = tf.nn.softplus(X)
            r = self.session.run(Y, {X: xi})
            np.testing.assert_almost_equal(r[0], [0.313261687, 0.693147181, 1.31326168], 7)

    def test_original_grad(self):
        X = tf.placeholder("float", [None, 3])
        for name in activations:
            Y = activations[name](X)
            grad = original_grad(Y.op, tf.ones_like(X))
            self.assertTrue('Tensor' in str(type(grad)))

    def test_warning_unsupported_activations(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
                X = tf.placeholder("float", [None, 3])
                Y = tf.nn.relu6(X)  # < an unsupported activation
                xi = [[-1, 0, 1]]
                de.explain('elrp', Y, X, xi)
                assert any(["unsupported activation" in str(wi.message) for wi in w])

    def test_override_as_default(self):
        """
        In DeepExplain context, nonlinearities behave as default, including training time
        """
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            r = train_xor(self.session)
            self.assertTrue(r)

    def test_explain_not_in_context(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            pass
        with self.assertRaises(RuntimeError) as cm:
            de.explain('grad*input', None, None, None)
        self.assertEqual(
            'Explain can be called only within a DeepExplain context.',
            str(cm.exception)
        )

    def test_invalid_method(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('invalid', None, None, None)
            self.assertIn('Method must be in',
                str(cm.exception)
            )

    # Failing on Python 2 on Travis !? But the warning is actually there
    # def test_gradient_was_not_overridden(self):
    #     X = tf.placeholder("float", [None, 3])
    #     Y = tf.nn.relu(X)
    #     with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
    #         with warnings.catch_warnings(record=True) as w:
    #             warnings.simplefilter("always")
    #             de.explain('grad*input', Y, X, [[0, 0, 0]])
    #             assert any(["DeepExplain detected" in str(wi.message) for wi in w])

    def test_T_is_tensor(self):
        X = tf.placeholder("float", [None, 3])
        Y = tf.nn.relu(X)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('grad*input', [Y], X, [[0, 0, 0]])
                self.assertIn('T must be a Tensorflow Tensor object',
                              str(cm.exception)
                              )

    def test_X_is_tensor(self):
        X = tf.placeholder("float", [None, 3])
        Y = tf.nn.relu(X)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('grad*input', Y, np.eye(3), [[0, 0, 0]])
                self.assertIn('Tensorflow Tensor object',
                              str(cm.exception)
                              )

    def test_all_in_X_are_tensor(self):
        X = tf.placeholder("float", [None, 3])
        Y = tf.nn.relu(X)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('grad*input', Y, [X, np.eye(3)], [[0, 0, 0]])
                self.assertIn('Tensorflow Tensor object',
                              str(cm.exception)
                              )

    def test_X_has_compatible_batch_dim(self):
        X = tf.placeholder("float", [10, 3])
        Y = tf.nn.relu(X)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('grad*input', Y, X, [[0, 0, 0]], batch_size=2)
                self.assertIn('the first dimension of the input tensor',
                              str(cm.exception)
                              )

    def test_T_has_compatible_batch_dim(self):
        X, out = simpler_model(self.session)
        xi = np.array([[-10, -5]]).repeat(50, 0)
        Y = out * np.expand_dims(np.array(range(50)), -1)
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            with self.assertRaises(RuntimeError) as cm:
                de.explain('saliency', Y, X, xi, batch_size=10)
                self.assertIn('the first dimension of the target tensor',
                              str(cm.exception)
                              )

    def test_use_of_target_weights(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, T = simple_model(tf.identity, self.session)
            xi = np.array([[1, 0]])
            yi1 = np.array([[1, 0]])
            yi2 = np.array([[0, 1]])
            yi3 = np.array([[1, 1]])
            yi4 = np.array([[0, 0]])

            a1 = de.explain('saliency', T, X, xi, ys=yi1)
            a2 = de.explain('saliency', T, X, xi, ys=yi2)
            a3 = de.explain('saliency', T, X, xi, ys=yi3)
            a4 = de.explain('saliency', T, X, xi, ys=yi4)
            np.testing.assert_almost_equal(a1+a2, a3, 10)
            np.testing.assert_almost_equal(a4, np.array([[0.0, 0.0]]), 10)

    def test_use_of_target_weights_batch(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, T = simple_model(tf.identity, self.session)
            xi = np.array([[1, 0]]).repeat(20, 0)
            yi1 = np.array([[1, 0]]).repeat(20, 0)
            yi2 = np.array([[0, 1]]).repeat(20, 0)
            yi3 = np.array([[1, 1]]).repeat(20, 0)
            yi4 = np.array([[0, 0]]).repeat(20, 0)

            a1 = de.explain('saliency', T, X, xi, ys=yi1, batch_size=5)
            a2 = de.explain('saliency', T, X, xi, ys=yi2, batch_size=5)
            a3 = de.explain('saliency', T, X, xi, ys=yi3, batch_size=5)
            a4 = de.explain('saliency', T, X, xi, ys=yi4, batch_size=5)
            np.testing.assert_almost_equal(a1+a2, a3, 10)
            np.testing.assert_almost_equal(a4, np.array([[0.0, 0.0]]).repeat(20, 0), 10)

    def test_wrong_weight_len(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, T = simple_model(tf.identity, self.session)
            xi = np.array([[1, 0]]).repeat(20, 0)
            yi1 = np.array([[1, 0]]) # < not same len as xi

            with self.assertRaises(RuntimeError) as cm:
                de.explain('saliency', T, X, xi, ys=yi1, batch_size=5)
                self.assertIn('the number of elements in ys must equal ',
                              str(cm.exception)
                              )

    def test_explainer_api_memory(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, T = simple_model(tf.identity, self.session)
            xi = np.array([[1, 0]]).repeat(20, 0)
            prev_ops_count = None
            explainer = de.get_explainer('saliency', T, X)
            for i in range(10):
                explainer.run(xi)
                # de.explain('saliency', T, X, xi)  < this would fail instead
                ops_count = len([n.name for n in tf.get_default_graph().as_graph_def().node])
                if prev_ops_count is None:
                    prev_ops_count = ops_count
                else:
                    self.assertEquals(prev_ops_count, ops_count)


class TestDummyMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_dummy_zero(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.sigmoid, self.session)
            xi = np.array([[10, -10]])
            attributions = de.explain('zero', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions[0], [0.0, 0.0], 10)

    def test_gradient_restored(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.sigmoid, self.session)
            xi = np.array([[10, -10]])
            de.explain('zero', out, X, xi)
            r = train_xor(self.session)
            self.assertTrue(r)


class TestSaliencyMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_saliency_method(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('saliency', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [1.0, 1.0]], 10)

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            attributions = de.explain('saliency', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[2.0, 2.0]], 10)

    def test_multiple_inputs_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            explainer = de.get_explainer('saliency', out, [X1, X2])
            attributions = explainer.run(xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[2.0, 2.0]], 10)


class TestGradInputMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_saliency_method(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('grad*input', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [3.0, -1.0]], 10)

    def test_saliency_method_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[-10, -5], [3, 1]])
            explainer = de.get_explainer('grad*input', out, X)
            attributions = explainer.run(xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [3.0, -1.0]], 10)

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            attributions = de.explain('grad*input', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[6.0, 2.0]], 10)


class TestIntegratedGradientsMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_int_grad(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('intgrad', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [1.5, -0.5]], 10)

    def test_int_grad_higher_precision(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('intgrad', out, X, xi, steps=500)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [1.5, -0.5]], 10)

    def test_int_grad_baseline(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[2, 0]])
            attributions = de.explain('intgrad', out, X, xi, baseline=np.array([1, 0]))
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[1.0, 0.0]], 10)

    def test_int_grad_baseline_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[2, 0]])
            explainer = de.get_explainer('intgrad', out, X, baseline=np.array([1, 0]))
            attributions = explainer.run(xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[1.0, 0.0]], 10)

    def test_int_grad_baseline_2(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[2, 0], [3, 0]])
            attributions = de.explain('intgrad', out, X, xi, baseline=np.array([1, 0]))
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[1.0, 0.0], [2.0, 0.0]], 10)

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            attributions = de.explain('intgrad', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[6.0, 2.0]], 10)

    def test_multiple_inputs_different_sizes(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model2(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3]])]
            attributions = de.explain('intgrad', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[6]], 10)

    def test_intgrad_targeting_equivalence(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.relu, self.session)
            xi = np.array([[5, 3]])
            self.assertEqual(out.shape[1], 2)
            a1 = de.explain('intgrad', out * np.array([[1, 0]]), X, xi)
            b1 = de.explain('intgrad', out * np.array([[0, 1]]), X, xi)
            a2 = de.explain('intgrad', out, X, xi, ys=np.array([[1, 0]]))
            b2 = de.explain('intgrad', out, X, xi, ys=np.array([[0, 1]]))
            np.testing.assert_almost_equal(a1, a2, 1)
            np.testing.assert_almost_equal(b1, b2, 1)


class TestEpsilonLRPMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_elrp_method(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('elrp', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [3.0, -1.0]], 3)

    def test_elrp_epsilon(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('elrp', out, X, xi, epsilon=1e-9)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [3.0, -1.0]], 7)

    def test_elrp_epsilon_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            explainer = de.get_explainer('elrp', out, X, epsilon=1e-9)
            attributions = explainer.run(xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [3.0, -1.0]], 7)

    def test_elrp_zero_epsilon(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            with self.assertRaises(AssertionError):
                de.explain('elrp', out, X, xi, epsilon=0)

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            attributions = de.explain('elrp', out, [X1, X2], xi, epsilon=1e-9)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 7)
            np.testing.assert_almost_equal(attributions[1], [[6.0, 2.0]], 7)

    def test_elrp_targeting_equivalence(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.relu, self.session)
            xi = np.array([[5, 3]])
            self.assertEqual(out.shape[1], 2)
            a1 = de.explain('elrp', out * np.array([[1, 0]]), X, xi)
            b1 = de.explain('elrp', out * np.array([[0, 1]]), X, xi)
            a2 = de.explain('elrp', out, X, xi, ys=np.array([[1, 0]]))
            b2 = de.explain('elrp', out, X, xi, ys=np.array([[0, 1]]))
            np.testing.assert_almost_equal(a1, a2, 1)
            np.testing.assert_almost_equal(b1, b2, 1)


class TestDeepLIFTMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_deeplift(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('deeplift', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [2.0, -1.0]], 10)

    def test_deeplift_batches(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            xi = np.repeat(xi, 25, 0)
            self.assertEqual(xi.shape[0], 50)
            attributions = de.explain('deeplift', out, X, xi, batch_size=32)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, np.repeat([[0.0, 0.0], [2.0, -1.0]], 25, 0), 10)

    def test_deeplift_batches_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            xi = np.repeat(xi, 25, 0)
            self.assertEqual(xi.shape[0], 50)
            explaoiner = de.get_explainer('deeplift', out, X)
            attributions = explaoiner.run(xi, batch_size=32)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, np.repeat([[0.0, 0.0], [2.0, -1.0]], 25, 0), 10)

    def test_deeplift_baseline(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[3, 1]])
            attributions = de.explain('deeplift', out, X, xi, baseline=xi[0])
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0]], 5)

    def test_multiple_inputs(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            attributions = de.explain('deeplift', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 7)
            np.testing.assert_almost_equal(attributions[1], [[6.0, 2.0]], 7)

    def test_multiple_inputs_different_sizes(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model2(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3]])]
            attributions = de.explain('deeplift', out, [X1, X2], xi)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], [[0.0, 0.0]], 10)
            np.testing.assert_almost_equal(attributions[1], [[6]], 10)

    def test_multiple_inputs_different_sizes_batches(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model2(self.session)
            xi = [np.array([[-10, -5]]).repeat(50, 0), np.array([[3]]).repeat(50, 0)]
            attributions = de.explain('deeplift', out, [X1, X2], xi, batch_size=32)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], np.repeat([[0.0, 0.0]], 50, 0), 10)
            np.testing.assert_almost_equal(attributions[1], np.repeat([[6]], 50, 0), 10)

    def test_multiple_inputs_different_sizes_batches_disable(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model2(self.session)
            xi = [np.array([[-10, -5]]).repeat(50, 0), np.array([[3]]).repeat(50, 0)]
            attributions = de.explain('deeplift', out, [X1, X2], xi, batch_size=None)
            self.assertEqual(len(attributions), len(xi))
            np.testing.assert_almost_equal(attributions[0], np.repeat([[0.0, 0.0]], 50, 0), 10)
            np.testing.assert_almost_equal(attributions[1], np.repeat([[6]], 50, 0), 10)

    def test_deeplift_targeting_equivalence(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.relu, self.session)
            xi = np.array([[5, 3]])
            self.assertEqual(out.shape[1], 2)
            a1 = de.explain('deeplift', out * np.array([[1, 0]]), X, xi)
            b1 = de.explain('deeplift', out * np.array([[0, 1]]), X, xi)
            a2 = de.explain('deeplift', out, X, xi, ys=np.array([[1, 0]]))
            b2 = de.explain('deeplift', out, X, xi, ys=np.array([[0, 1]]))
            np.testing.assert_almost_equal(a1, a2, 1)
            np.testing.assert_almost_equal(b1, b2, 1)


class TestOcclusionMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_occlusion(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('occlusion', out, X, xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [1.0, -1.0]], 10)

    def test_occlusion_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]])
            explainer = de.get_explainer('occlusion', out, X)
            attributions = explainer.run(xi)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [1.0, -1.0]], 10)

    def test_occlusion_batches(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model( self.session)
            xi = np.array([[-10, -5], [3, 1]]).repeat(10, 0)
            attributions = de.explain('occlusion', out, X, xi, batch_size=5)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, np.repeat([[0.0, 0.0], [1.0, -1.0]], 10, 0), 10)

    def test_window_shape(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simpler_model(self.session)
            xi = np.array([[-10, -5], [3, 1]])
            attributions = de.explain('occlusion', out, X, xi, window_shape=(2,))
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, 0.0], [0.5, 0.5]], 10)

    def test_nan_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
                X, out = simpler_model(self.session)
                xi = np.array([[-10, -5], [3, 1]])
                attributions = de.explain('occlusion', out, X, xi, step=2)
                self.assertEqual(attributions.shape, xi.shape)
                np.testing.assert_almost_equal(attributions, [[0.0, np.nan], [1.0, np.nan]], 10)
                assert any(["nans" in str(wi.message) for wi in w])

    def test_multiple_inputs_error(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            with self.assertRaises(RuntimeError) as cm:
                de.explain('occlusion', out, [X1, X2], xi)
            self.assertIn('not yet supported', str(cm.exception))

    def test_occlusion_targeting_equivalence(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.nn.relu, self.session)
            xi = np.array([[5, 3]])
            self.assertEqual(out.shape[1], 2)
            a1 = de.explain('occlusion', out * np.array([[1, 0]]), X, xi)
            b1 = de.explain('occlusion', out * np.array([[0, 1]]), X, xi)
            a2 = de.explain('occlusion', out, X, xi, ys=np.array([[1, 0]]))
            b2 = de.explain('occlusion', out, X, xi, ys=np.array([[0, 1]]))
            np.testing.assert_almost_equal(a1, a2, 10)
            np.testing.assert_almost_equal(b1, b2, 10)


class TestShapleySamplingMethod(TestCase):

    def setUp(self):
        self.session = tf.Session()

    def tearDown(self):
        self.session.close()
        tf.reset_default_graph()

    def test_shapley_sampling(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = min_model(self.session)
            xi = np.array([[2, -2], [4, 2]])
            attributions = de.explain('shapley_sampling', out, X, xi, samples=300)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, [[0.0, -2.0], [1.0, 1.0]], 1)

    def test_shapley_sampling_batches(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = min_model(self.session)
            xi = np.array([[2, -2], [4, 2]]).repeat(20, 0)
            attributions = de.explain('shapley_sampling', out, X, xi, samples=300, batch_size=5)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, np.repeat([[0.0, -2.0], [1.0, 1.0]], 20, 0), 1)

    def test_shapley_sampling_batches_explainer_api(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = min_model(self.session)
            xi = np.array([[2, -2], [4, 2]]).repeat(20, 0)
            explainer = de.get_explainer('shapley_sampling', out, X, samples=300)
            attributions = explainer.run(xi, batch_size=5)
            self.assertEqual(attributions.shape, xi.shape)
            np.testing.assert_almost_equal(attributions, np.repeat([[0.0, -2.0], [1.0, 1.0]], 20, 0), 1)

    def test_shapley_sampling_dims(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = min_model_2d(self.session)
            xi = np.array([[[-1, 4], [-2, 1]]])
            attributions = de.explain('shapley_sampling', out, X, xi, samples=300, sampling_dims=[1])
            self.assertEqual(attributions.shape, (1, 2))
            np.testing.assert_almost_equal(attributions, [[-.5, -1.5]], 1)

    def test_multiple_inputs_error(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X1, X2, out = simple_multi_inputs_model(self.session)
            xi = [np.array([[-10, -5]]), np.array([[3, 1]])]
            with self.assertRaises(RuntimeError) as cm:
                de.explain('shapley_sampling', out, [X1, X2], xi)
            self.assertIn('not yet supported', str(cm.exception))

    def test_shapley_targeting_equivalence(self):
        with DeepExplain(graph=tf.get_default_graph(), session=self.session) as de:
            X, out = simple_model(tf.identity, self.session)
            xi = np.array([[5, 3]])
            self.assertEqual(out.shape[1], 2)
            np.random.seed(10)
            a1 = de.explain('shapley_sampling', out * np.array([[1, 0]]), X, xi, samples=10)
            np.random.seed(10)
            b1 = de.explain('shapley_sampling', out * np.array([[0, 1]]), X, xi, samples=10)
            np.random.seed(10)
            a2 = de.explain('shapley_sampling', out, X, xi, ys=np.array([[1, 0]]), samples=10)
            np.random.seed(10)
            b2 = de.explain('shapley_sampling', out, X, xi, ys=np.array([[0, 1]]), samples=10)
            np.testing.assert_almost_equal(a1, a2, 3)
            np.testing.assert_almost_equal(b1, b2, 3)


"""
Unit testing second order matrix vector product.
"""

from __future__ import (division, print_function, unicode_literals)

import numpy as np
import tensorflow as tf

from tensorflow_forward_ad import logger
from tensorflow_forward_ad.fwgrad import forward_gradients
from tensorflow_forward_ad.second_order import (
    hessian_vec_fw, hessian_vec_bk, gauss_newton_vec, fisher_vec_fw,
    fisher_vec_bk, gauss_newton_vec_z, fisher_vec_z)

log = logger.get()


class TestSecondOrderFwGrad(tf.test.TestCase):

  def test_hessian_quadratic(self):
    rnd = np.random.RandomState(0)
    dtype = tf.float64
    with tf.Graph().as_default():
      r = tf.Variable(0.0, dtype=dtype)
      x = tf.constant(rnd.uniform(-1.0, 1.0, [2, 27]), dtype=dtype, name="x")
      w2 = tf.constant(rnd.uniform(-1.0, 1.0, [27, 1]), dtype=dtype, name="w2")
      v2 = tf.constant(rnd.uniform(-1.0, 1.0, [27, 1]), dtype=dtype, name="v2")
      w2v = tf.add(w2, tf.multiply(r, v2))
      h2 = tf.matmul(x, w2v)
      y2 = tf.reduce_sum(h2 * h2)

      grad_w = tf.gradients(y2, w2)
      hv_fw = hessian_vec_fw(y2, [w2v], [v2])
      hv_bk = hessian_vec_bk(y2, [w2], [v2])

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        grad_w = sess.run(grad_w)
        hv_fw_val = sess.run(hv_fw)
        hv_bk_val = sess.run(hv_bk)
        np.testing.assert_allclose(hv_fw_val, hv_bk_val, rtol=1e-5)

  def test_sparse_softmax_with_logits_grad(self):
    rnd = np.random.RandomState(0)
    dtype = tf.float64  # Use float64 and CPU for finite difference checking.
    with tf.Graph().as_default(), tf.device("/cpu:0"):
      r = tf.Variable(0.0, dtype=dtype)
      # Input.
      x = tf.constant(rnd.uniform(-1.0, 1.0, [2, 27]), dtype=dtype, name="x")

      # Fully connected.
      v = tf.constant(rnd.uniform(-0.1, 0.1, [27, 2]), dtype=dtype, name="v")
      w = tf.constant(rnd.uniform(-1.0, 1.0, [27, 2]), dtype=dtype, name="w")
      y = tf.matmul(x, w + r * v)

      label = tf.constant([1, 0], dtype=tf.int32)
      loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
          logits=y, labels=label)
      loss = tf.reduce_sum(loss)

      hv_fw = hessian_vec_fw(loss, [w], [v])[0]
      # hv_bk = hessian_vec_bk(loss, [w], [v])[0]
      grad_w = tf.gradients(loss, [w])[0]

      with self.test_session() as sess:
        # Compute Hv with finite difference.
        sess.run(tf.global_variables_initializer())
        eps = 1e-5
        sess.run(tf.assign(r, -eps))
        grad_w_val1 = sess.run(grad_w)
        sess.run(tf.assign(r, eps))
        grad_w_val2 = sess.run(grad_w)
        fndv = (grad_w_val2 - grad_w_val1) / (2 * eps)

        # Compute Hv with forward mode autodiff.
        sess.run(tf.global_variables_initializer())
        fwv = sess.run(hv_fw)

        # # Compute Hv with reverse mode autodiff.
        # bkv = sess.run(hv_bk)

        np.testing.assert_allclose(fndv, fwv, rtol=1e-5)
        # # Expect failure here.
        #np.testing.assert_allclose(fndv, bkv, rtol=1e-5)

  def test_softmax_grad(self):
    rnd = np.random.RandomState(0)
    dtype = tf.float64  # Use float64 and CPU for finite difference checking.
    with tf.Graph().as_default(), tf.device("/cpu:0"):
      r = tf.Variable(0.0, dtype=dtype)
      # Input.
      x = tf.constant(rnd.uniform(-1.0, 1.0, [2, 27]), dtype=dtype, name="x")

      # Fully connected.
      v = tf.constant(rnd.uniform(-0.1, 0.1, [27, 2]), dtype=dtype, name="v")
      w = tf.constant(rnd.uniform(-1.0, 1.0, [27, 2]), dtype=dtype, name="w")
      y = tf.matmul(x, w + r * v)

      label = tf.constant([1, 0], dtype=tf.int32)
      y = tf.nn.softmax(y)
      t = tf.one_hot(label, 2, dtype=dtype)
      loss = tf.log(y + 1e-5) * t
      loss = tf.reduce_sum(loss)

      hv_fw = hessian_vec_fw(loss, [w], [v])[0]
      hv_bk = hessian_vec_bk(loss, [w], [v])[0]
      grad_w = tf.gradients(loss, [w])[0]

      with self.test_session() as sess:
        # Compute Hv with finite difference.
        sess.run(tf.global_variables_initializer())
        eps = 1e-5
        sess.run(tf.assign(r, -eps))
        grad_w_val1 = sess.run(grad_w)
        sess.run(tf.assign(r, eps))
        grad_w_val2 = sess.run(grad_w)
        fndv = (grad_w_val2 - grad_w_val1) / (2 * eps)

        # Compute Hv with forward mode autodiff.
        sess.run(tf.global_variables_initializer())
        fwv = sess.run(hv_fw)

        # Compute Hv with reverse mode autodiff.
        bkv = sess.run(hv_bk)

        np.testing.assert_allclose(fndv, fwv, rtol=1e-5)
        np.testing.assert_allclose(fndv, bkv, rtol=1e-5)
        np.testing.assert_allclose(fwv, bkv, rtol=1e-5)

  def _test_hessian_cnn(self):
    rnd = np.random.RandomState(0)
    dtype = tf.float32  # Use float64 and CPU for finite difference checking.
    # tf.nn.conv2d and tf.nn.max_pool does not support float64.
    # with tf.Graph().as_default(), tf.device("/cpu:0"):
    with tf.Graph().as_default():
      # Input.
      x = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 5, 5, 2]), dtype=dtype, name="x")

      # First convolution.
      v = tf.constant(
          np.random.uniform(-0.1, 0.1, [2, 2, 2, 3]), dtype=dtype, name="v")
      w = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 2, 3]), dtype=dtype, name="w")
      h = tf.nn.conv2d(x, w, [1, 1, 1, 1], "SAME")
      h = tf.nn.max_pool(h, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")
      h = tf.nn.relu(h)

      # Second convolution.
      v_ = tf.constant(
          np.random.uniform(-0.1, 0.1, [2, 2, 3, 3]), dtype=dtype, name="v_")
      w_ = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 3, 3]), dtype=dtype, name="w_")
      h = tf.nn.conv2d(h, w_, [1, 1, 1, 1], "SAME")
      h = tf.nn.sigmoid(h)

      # Fully connected.
      dim = 27
      v2 = tf.constant(rnd.uniform(-0.1, 0.1, [dim, 2]), dtype=dtype, name="v2")
      w2 = tf.constant(rnd.uniform(-1.0, 1.0, [dim, 2]), dtype=dtype, name="w2")
      w2v = w2 + r * v2
      h = tf.reshape(h, [-1, dim])
      y = tf.matmul(h, w2v)

      label = tf.constant([1, 0], dtype=tf.int32)
      # loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
      #     logits=y, labels=label)
      # loss = tf.reduce_sum(loss)

      # Use manual cross entropy.
      label_dense = tf.one_hot(label, 2, dtype=dtype)
      y = tf.nn.softmax(y)
      loss = label_dense * tf.log(y + 1e-5)
      loss = tf.reduce_sum(loss)

      # Using explicit R-op is deprecated, due to complex model building.
      hv_fw = hessian_vec_fw(loss, [w2v, w_v, wv], [v2, v_, v])
      hv_bk = hessian_vec_bk(loss, [w2, w_, w], [v2, v_, v])

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        hv_fw_val = sess.run(hv_fw)
        hv_bk_val = sess.run(hv_bk)
        for fwv, bkv in zip(hv_fw_val, hv_bk_val):
          np.testing.assert_allclose(fwv, bkv, rtol=1e-5)

  def test_gauss_newton_quadratic(self):
    rnd = np.random.RandomState(0)
    with tf.Graph().as_default():
      r = tf.Variable(0.0)
      x = tf.constant(
          rnd.uniform(-1.0, 1.0, [2, 27]), dtype=tf.float32, name="x")
      w = tf.constant(
          rnd.uniform(-1.0, 1.0, [27, 3]), dtype=tf.float32, name="w2")
      v = tf.constant(
          rnd.uniform(-1.0, 1.0, [27, 3]), dtype=tf.float32, name="v2")
      wv = tf.add(w, tf.multiply(r, v))
      z = tf.matmul(x, wv)
      y = 0.5 * tf.reduce_sum(z * z)

      # Gauss-Newton, same as Fisher for square loss.
      gv_fw = gauss_newton_vec(y, z, [w], [v])[0]
      # Fisher towards the output layer.
      fv_fw = fisher_vec_fw(z, [w], [v])

      # Fisher using tf.gradients (reverse mode).
      fv_bk = fisher_vec_bk(y, [w], [v])
      # Fisher using forward mode, towards loss function.
      fv_fw_y = fisher_vec_fw(y, [w], [v])

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())

        gv_fw_val = sess.run(gv_fw)
        fv_fw_val = sess.run(fv_fw)
        np.testing.assert_allclose(gv_fw_val, fv_fw_val, rtol=1e-5, atol=1e-5)

        fv_fw_y_val = sess.run(fv_fw_y)
        fv_bk_val = sess.run(fv_bk)
        np.testing.assert_allclose(fv_fw_y_val, fv_bk_val, rtol=1e-5, atol=1e-5)

  def test_gauss_newton_output_quadratic(self):
    rnd = np.random.RandomState(0)
    with tf.Graph().as_default():
      x = tf.constant(
          rnd.uniform(-1.0, 1.0, [27, 2]), dtype=tf.float32, name="x")
      h = tf.constant(
          rnd.uniform(-1.0, 1.0, [3 * 2, 3 * 2]), dtype=tf.float32, name="h")
      h = tf.matmul(tf.transpose(h), h)
      j = tf.constant(
          rnd.uniform(-1.0, 1.0, [3, 27]), dtype=tf.float32, name="j")
      z = tf.matmul(j, x)  # [3, 2]
      z_ = tf.reshape(z, [3 * 2, 1])  # [6, 1]
      y = 0.5 * tf.matmul(tf.matmul(tf.transpose(z_), h), z_)  # [1, 1]
      v = tf.constant(
          rnd.uniform(-1.0, 1.0, [3, 2]), dtype=tf.float32, name="v")
      act_jv = tf.gradients(z, x, v, gate_gradients=True)[0]
      act_hjjv = gauss_newton_vec_z(y, z, x, v)[0]
      exp_jv = tf.matmul(tf.transpose(j), v)  # [27, 2]
      exp_jjv = tf.matmul(j, exp_jv)  # [3, 2]
      exp_jjv_ = tf.reshape(exp_jjv, [3 * 2, 1])
      exp_hjjv = tf.matmul(h, exp_jjv_)  # [6, 1]
      exp_hjjv = tf.reshape(exp_hjjv, [3, 2])

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        h_val = sess.run(h)
        e, _ = np.linalg.eig(h_val)
        self.assertTrue((e > 0).all())
        act_hjjv_val, exp_hjjv_val, act_jv_val, exp_jv_val = sess.run(
            [act_hjjv, exp_hjjv, act_jv, exp_jv])
        np.testing.assert_allclose(act_jv_val, exp_jv_val, rtol=1e-5)
        np.testing.assert_allclose(act_hjjv_val, exp_hjjv_val, rtol=1e-5)

  def test_fisher_output_quadratic(self):
    rnd = np.random.RandomState(0)
    with tf.Graph().as_default():
      x = tf.constant(
          rnd.uniform(-1.0, 1.0, [27, 2]), dtype=tf.float32, name="x")
      h = tf.constant(
          rnd.uniform(-1.0, 1.0, [3 * 2, 3 * 2]), dtype=tf.float32, name="h")
      h = tf.matmul(tf.transpose(h), h)
      j = tf.constant(
          rnd.uniform(-1.0, 1.0, [3, 27]), dtype=tf.float32, name="j")
      z = tf.matmul(j, x)  # [3, 2]
      z_ = tf.reshape(z, [3 * 2, 1])  # [6, 1]
      y = 0.5 * tf.matmul(tf.matmul(tf.transpose(z_), h), z_)  # [1, 1]
      v = tf.constant(
          rnd.uniform(-1.0, 1.0, [3, 2]), dtype=tf.float32, name="v")
      act_jjv = fisher_vec_z(z, x, v)[0]
      exp_jv = tf.matmul(tf.transpose(j), v)  # [27, 2]
      exp_jjv = tf.matmul(j, exp_jv)  # [3, 2]

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        act_jjv_val, exp_jjv_val = sess.run([act_jjv, exp_jjv])
        np.testing.assert_allclose(act_jjv_val, exp_jjv_val, rtol=1e-5)


if __name__ == "__main__":
  tf.test.main()



"""
Unit testing forward-mode automatic differentiation.
"""

from __future__ import (division, print_function, unicode_literals)

import numpy as np
import tensorflow as tf

from tensorflow_forward_ad import logger
from tensorflow_forward_ad.fwgrad import forward_gradients

log = logger.get()
tf.logging.set_verbosity(tf.logging.ERROR)


class BasicFwGradTests(tf.test.TestCase):

  def inner_prod(self, xlist, ylist):
    """Computes inner product of two lists of tensors."""
    return tf.reduce_sum(
        tf.stack([tf.reduce_sum(x * y) for x, y in zip(xlist, ylist)]))

  def make_unit_graph(self, x, y, rnd=None, dtype=tf.float32):
    """Makes a computation graph that computes (J^T r)^T v and r^T J v"""
    if rnd is None:
      rnd = np.random.RandomState(0)
    x_shape = [int(ss) for ss in x.get_shape()]
    v = self.get_random_tensor(x_shape, rnd=rnd)
    y_shape = [int(ss) for ss in y.get_shape()]
    r = self.get_random_tensor(y_shape, rnd=rnd)
    jt_r = tf.gradients(y, [x], r, gate_gradients=True)
    jt_r_t_v = self.inner_prod(jt_r, [v])
    j_v = forward_gradients(y, [x], [v], gate_gradients=True)
    r_t_j_v = tf.reduce_sum(r * j_v)
    return jt_r_t_v, r_t_j_v

  def assert_bw_fw(self, sess, x, y, rnd=None):
    bk, fw = self.make_unit_graph(x, y, rnd=rnd)
    bk_val, fw_val = sess.run([bk, fw])
    np.testing.assert_allclose(bk_val, fw_val, rtol=1e-5)

  def get_random_tensor(self,
                        shape,
                        rnd=None,
                        dtype=tf.float32,
                        scale=(-1.0, 1.0)):
    return tf.constant(rnd.uniform(scale[0], scale[1], shape), dtype=dtype)


class MaxPool_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 5, 5, 3], rnd=rnd)
      y = tf.nn.max_pool(x, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class MatMul_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([12, 24], rnd=rnd)
      y = tf.matmul(x, x2)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Conv2D_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 5, 5, 3], rnd=rnd)
      x2 = tf.constant(rnd.uniform(-1.0, 1.0, [3, 3, 3, 4]), dtype=tf.float32)
      y = tf.nn.conv2d(x, x2, [1, 1, 1, 1], "SAME")
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic2(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 5, 5, 3], rnd=rnd)
      x2 = tf.constant(rnd.uniform(-1.0, 1.0, [3, 3, 3, 4]), dtype=tf.float32)
      y = tf.nn.conv2d(x, x2, [1, 1, 1, 1], "VALID")
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Conv2DBackpropFilter_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 5, 5, 3], rnd=rnd)
      x2 = tf.constant(rnd.uniform(-1.0, 1.0, [3, 3, 3, 4]), dtype=tf.float32)
      y = tf.gradients(tf.nn.conv2d(x, x2, [1, 1, 1, 1], "SAME"), x2)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Conv2DBackpropInput_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([5, 5, 4, 3], rnd=rnd)
      x2 = tf.constant(rnd.uniform(-1.0, 1.0, [3, 3, 3, 4]), dtype=tf.float32)
      y = tf.gradients(tf.nn.conv2d(x2, x, [1, 1, 1, 1], "SAME"), x2)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class MaxPoolGrad_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 5, 5, 3], rnd=rnd)
      y = tf.gradients(
          tf.square(tf.nn.max_pool(x, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")),
          x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Square_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      dtype = tf.float32
      x_shape = [18, 12]
      x = tf.constant(rnd.uniform(-1.0, 1.0, x_shape), dtype=dtype, name="x")
      y = tf.square(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Sqrt_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      dtype = tf.float32
      x_shape = [18, 12]
      x = tf.constant(rnd.uniform(0.0, 1.0, x_shape), dtype=dtype, name="x")
      y = tf.sqrt(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Relu_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      dtype = tf.float32
      x_shape = [18, 12]
      x = tf.constant(rnd.uniform(-1.0, 1.0, x_shape), dtype=dtype, name="x")
      y = tf.nn.relu(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class ReluGrad_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.square(tf.nn.relu(x)), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Maximum_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.maximum(x, x2)
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Minimum_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.minimum(x, x2)
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Add_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = x + x2
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Sub_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = x - x2
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Neg_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.negative(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class AddN_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.add_n([x, x2])
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Mul_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = x * x2
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Div_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = x / x2
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Reciprocal_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reciprocal(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Pow_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.pow(x, x2)
      self.assert_bw_fw(sess, x, y, rnd=rnd)
      self.assert_bw_fw(sess, x2, y, rnd=rnd)


class Identity_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.identity(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Print_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.Print(x, [x])
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Sigmoid_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.sigmoid(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class SigmoidGrad_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.sigmoid(x), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic2(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.square(tf.sigmoid(x)), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Tanh_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.tanh(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)

    def test_manual(self):
      with tf.Graph().as_default(), tf.device("/cpu:0"):
        with self.test_session() as sess:
          x_val = np.random.uniform(0, 1)
          x = tf.constant(x_val)
          y = tf.tanh(x)
          dy_dx = forward_gradients(y, x, gate_gradients=True)
          dy_dx_tf = sess.run(dy_dx)
          eps = 1e-5
          x_val = x_val - eps
          y_val_1 = np.tanh(x_val)
          x_val = x_val + 2 * eps
          y_val_2 = np.tanh(x_val)
          dy_dx_fd = (y_val_2 - y_val_1) / (2 * eps)
          np.testing.assert_allclose(dy_dx_tf, dy_dx_fd, rtol=1e-5)


class TanhGrad_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.tanh(x), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic2(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.square(tf.tanh(x)), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Softmax_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.nn.softmax(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class SoftmaxGrad_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.gradients(tf.square(tf.nn.softmax(x)), x)[0]
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Log_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd, scale=(0.1, 10.0))
      y = tf.log(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class SparseSoftmaxCrossEntropyWithLogits_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      labels = tf.constant(
          rnd.uniform(0, 5, [18]).astype(np.int64), dtype=tf.int64)
      if tf.__version__.startswith("0"):
        y = tf.nn.sparse_softmax_cross_entropy_with_logits(x, labels)
      else:
        y = tf.nn.sparse_softmax_cross_entropy_with_logits(
            logits=x, labels=labels)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Transpose_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.transpose(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Pack_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      if tf.__version__.startswith("0"):
        y = tf.pack([x, x2])
      else:
        y = tf.stack([x, x2])
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Concat_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      for ax in [0, 1]:
        if tf.__version__.startswith("0"):
          y = tf.concat(ax, [x, x2])
        else:
          y = tf.concat([x, x2], axis=ax)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Pack_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      x2 = self.get_random_tensor([18, 12], rnd=rnd)
      if tf.__version__.startswith("0"):
        y = tf.pack([x, x2])
      else:
        y = tf.stack([x, x2])
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Reshape_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reshape(x, [4, -1])
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class ReduceSum_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_sum(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic2(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_sum(x, [1])
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic3(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_sum(x, [1], keep_dims=True)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class ReduceMean_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_mean(x)
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic2(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_mean(x, [1])
      self.assert_bw_fw(sess, x, y, rnd=rnd)

  def test_basic3(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([18, 12], rnd=rnd)
      y = tf.reduce_mean(x, [1], keep_dims=True)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class Tile_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 12], rnd=rnd)
      y = tf.tile(x, [4, 1])
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class ExpandDims_FwGradTests(BasicFwGradTests):

  def test_basic(self):
    with tf.Graph().as_default(), self.test_session() as sess:
      rnd = np.random.RandomState(0)
      x = self.get_random_tensor([2, 12], rnd=rnd)
      y = tf.expand_dims(x, 0)
      self.assert_bw_fw(sess, x, y, rnd=rnd)


class ForwardGradGraphTests(tf.test.TestCase):

  def test_convnet(self):
    with tf.Graph().as_default():
      # Define model.
      r = tf.Variable(1.0)
      x = tf.constant(
          np.random.uniform(-1.0, 1.0, [1, 5, 5, 2]), dtype=tf.float32)
      w = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 2, 3]), dtype=tf.float32)
      h = tf.nn.conv2d(r + x, r * w, [1, 1, 1, 1], "SAME")
      h = tf.nn.max_pool(h, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")
      h = tf.nn.relu(h)
      # First branch.
      w2 = tf.constant(np.random.uniform(-1.0, 1.0, [27, 1]), dtype=tf.float32)
      h2 = tf.matmul(tf.reshape(h, [1, -1]), w2)
      y2 = tf.nn.tanh(h2)
      y2 = tf.reduce_sum(y2)
      # We can take a second branch.
      w3 = tf.constant(np.random.uniform(-1.0, 1.0, [27, 1]), dtype=tf.float32)
      h3 = tf.matmul(tf.reshape(h, [1, -1]), w3)
      y3 = tf.nn.sigmoid(h3)
      y3 = tf.reduce_sum(y3)
      # Take gradients of a list of y wrt. scalar r.
      # Returns [grad_y2_r, grad_y3_r].
      grad_fw = forward_gradients([y2, y3], r, gate_gradients=True)
      # Reverse mode implementation from tensorflow.
      grad_bk = [tf.gradients(y2, r)[0], tf.gradients(y3, r)[0]]
      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        grad_fw_val = sess.run(grad_fw)
        grad_bk_val = sess.run(grad_bk)
        np.testing.assert_allclose(grad_fw_val, grad_bk_val, rtol=5)

  def test_grad_graph(self):
    with tf.Graph().as_default():

      # Dummy variable.
      r = tf.Variable(1.0)

      # Input.
      x = tf.constant(
          np.random.uniform(-1.0, 1.0, [1, 5, 5, 2]),
          dtype=tf.float32,
          name="x")

      # First convolution.
      v = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 2, 3]),
          dtype=tf.float32,
          name="v")
      w = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 2, 3]),
          dtype=tf.float32,
          name="w")
      wv = w + r * v
      h = tf.nn.conv2d(x, wv, [1, 1, 1, 1], "SAME")
      h = tf.nn.max_pool(h, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")
      h = tf.nn.relu(h)

      # Second convolution.
      v_ = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 3, 3]),
          dtype=tf.float32,
          name="v_")
      w_ = tf.constant(
          np.random.uniform(-1.0, 1.0, [2, 2, 3, 3]),
          dtype=tf.float32,
          name="w_")
      w_v = w_ + r * v_
      h = tf.nn.conv2d(h, w_v, [1, 1, 1, 1], "SAME")

      # Fully connected.
      w2 = tf.constant(
          np.random.uniform(-1.0, 1.0, [27, 1]), dtype=tf.float32, name="w2")
      h2 = tf.matmul(tf.reshape(h, [1, -1]), w2)
      y2 = tf.nn.sigmoid(h2)
      y2 = tf.reduce_sum(y2)
      grad_bk = tf.gradients(y2, [w, w_], gate_gradients=True)
      grad_fw = forward_gradients(grad_bk, r, gate_gradients=True)
      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        sess.run(grad_fw)

  def test_op(self):
    with tf.Graph().as_default():
      r = tf.Variable(1.0)
      f = tf.constant(0.0)
      r = tf.add_n([r, f])
      g = tf.get_default_graph()
      node_list = g.as_graph_def().node
      inspect = set(["AddN"])
      node_list = filter(lambda x: x.op in inspect, node_list)

  def test_op_2(self):
    rnd = np.random.RandomState(0)
    with tf.Graph().as_default():
      logits = tf.constant(rnd.uniform(0.0, 1.0, [2, 2]), dtype=tf.float32)
      logits_v = tf.constant(rnd.uniform(0.0, 1.0, [2, 2]), dtype=tf.float32)
      r = tf.Variable(0.0, dtype=tf.float32)
      logits = logits_v * r + logits
      t = tf.constant([1, 0])
      t = tf.one_hot(t, 2, dtype=tf.float32)
      # loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
      #     logits=logits, labels=y)
      y = tf.nn.softmax(logits)
      loss = t * tf.log(y + 1e-5)
      loss = tf.reduce_sum(loss)

      g = tf.get_default_graph()
      node_list = g.as_graph_def().node
      inspect = set(["Softmax", "SoftmaxGrad"])
      node_list = filter(lambda x: x.op in inspect, node_list)
      grad_fw = forward_gradients(loss, r, gate_gradients=True)
      grad_bk = tf.gradients(loss, r)[0]
      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        grad_bk_val = sess.run(grad_bk)
        grad_fw_val = sess.run(grad_fw)
        np.testing.assert_allclose(grad_fw_val, grad_bk_val, rtol=5)

  def test_forward_mode_cnn(self):
    """Test v^T (J v) = (J^T v) ^T v"""
    rnd = np.random.RandomState(0)
    dtype = tf.float32  # Use float64 and CPU for finite difference checking.
    # tf.nn.conv2d and tf.nn.max_pool does not support float64.
    # with tf.Graph().as_default(), tf.device("/cpu:0"):
    with tf.Graph().as_default():
      # Input.
      x = tf.constant(
          rnd.uniform(-1.0, 1.0, [2, 5, 5, 2]), dtype=dtype, name="x")

      # First convolution.
      v = tf.constant(
          rnd.uniform(-1.0, 1.0, [2, 2, 2, 3]), dtype=dtype, name="v")
      w = tf.constant(
          rnd.uniform(-1.0, 1.0, [2, 2, 2, 3]), dtype=dtype, name="w")
      h = tf.nn.conv2d(x, w, [1, 1, 1, 1], "SAME")
      h = tf.nn.max_pool(h, [1, 3, 3, 1], [1, 2, 2, 1], "SAME")
      h = tf.nn.relu(h)

      # Second convolution.
      v_ = tf.constant(
          rnd.uniform(-0.1, 0.1, [2, 2, 3, 3]), dtype=dtype, name="v_")
      w_ = tf.constant(
          rnd.uniform(-1.0, 1.0, [2, 2, 3, 3]), dtype=dtype, name="w_")
      h = tf.nn.conv2d(h, w_, [1, 1, 1, 1], "SAME")
      h = tf.nn.sigmoid(h)

      # Fully connected.
      dim = 27
      v2 = tf.constant(rnd.uniform(-0.1, 0.1, [dim, 2]), dtype=dtype, name="v2")
      w2 = tf.constant(rnd.uniform(-1.0, 1.0, [dim, 2]), dtype=dtype, name="w2")
      h = tf.reshape(h, [-1, dim])
      y = tf.matmul(h, w2)
      r = tf.constant(rnd.uniform(-1.0, 1.0, [2, 2]), dtype=dtype, name="r")

      w_list = [w, w_, w2]
      v_list = [v, v_, v2]

      # Taking inner product of two list of tensors.
      inner_prod = lambda xlist, ylist: tf.reduce_sum(
        tf.stack([tf.reduce_sum(x * y) for x, y in zip(xlist, ylist)]))

      # J^T r
      jt_r = tf.gradients(y, w_list, r, gate_gradients=True)
      # (J^T r)^T v
      jt_r_t_v = inner_prod(jt_r, v_list)

      # J v
      j_v = forward_gradients(y, w_list, v_list, gate_gradients=True)
      # r^T J v
      r_t_j_v = tf.reduce_sum(r * j_v)

      with self.test_session() as sess:
        sess.run(tf.global_variables_initializer())
        bk_val, fw_val = sess.run([jt_r_t_v, r_t_j_v])
        np.testing.assert_allclose(bk_val, fw_val, rtol=1e-5)


if __name__ == "__main__":
  tf.test.main()



import numpy as np
import tensorflow as tf

# Some of the following functions for batch processing have been borrowed and adapter from Keras
# https://github.com/keras-team/keras/blob/master/keras/utils/generic_utils.py
# https://github.com/keras-team/keras/blob/master/keras/engine/training_utils.py


def make_batches(size, batch_size):
    """Returns a list of batch indices (tuples of indices).
    # Arguments
        size: Integer, total size of the data to slice into batches.
        batch_size: Integer, batch size.
    # Returns
        A list of tuples of array indices.
    """
    num_batches = (size + batch_size - 1) // batch_size  # round up
    return [(i * batch_size, min(size, (i + 1) * batch_size))
            for i in range(num_batches)]


def to_list(x, allow_tuple=False):
    """Normalizes a list/tensor into a list.
    If a tensor is passed, we return
    a list of size 1 containing the tensor.
    # Arguments
        x: target object to be normalized.
        allow_tuple: If False and x is a tuple,
            it will be converted into a list
            with a single element (the tuple).
            Else converts the tuple to a list.
    # Returns
        A list.
    """
    if isinstance(x, list):
        return x
    if allow_tuple and isinstance(x, tuple):
        return list(x)
    return [x]


def unpack_singleton(x):
    """Gets the equivalent np-array if the iterable has only one value.
    Otherwise return the iterable.
    # Argument
        x: A list or tuple.
    # Returns
        The same iterable or the iterable converted to a np-array.
    """
    if len(x) == 1:
        return np.array(x)
    return x


def slice_arrays(arrays, start=None, stop=None):
    """Slices an array or list of arrays.
    """
    if arrays is None:
        return [None]
    elif isinstance(arrays, list):
        return [None if x is None else x[start:stop] for x in arrays]
    else:
        return arrays[start:stop]


def placeholder_from_data(numpy_array):
    if numpy_array is None:
        return None
    return tf.placeholder('float', [None,] + list(numpy_array.shape[1:]))


from __future__ import division, print_function

import numpy as np
import tensorflow as tf

from tensorflow.python.ops import gradient_checker
from tensorflow_forward_ad import maxpool_gradgrad


class MaxPoolGradGradTests(tf.test.TestCase):

  def test_basics(self):
    tf.set_random_seed(1234)
    x_shape = [4, 4, 4, 5]
    rnd = np.random.RandomState(0)
    x_np = rnd.uniform(-1.0, 1.0, x_shape).astype(np.float32)

    # test op max_pool_grad
    with tf.Graph().as_default(), tf.Session() as sess:
      x_tf = tf.constant(x_np, name="x")
      y_tf1, _ = tf.nn.max_pool_with_argmax(
          x_tf,
          ksize=[1, 3, 3, 1],
          strides=[1, 2, 2, 1],
          padding='SAME',
          name="y1")
      y_tf2 = tf.nn.max_pool(
          x_tf,
          ksize=[1, 3, 3, 1],
          strides=[1, 2, 2, 1],
          padding='SAME',
          name="y2")
      z_tf1 = tf.reduce_sum(tf.square(y_tf1))
      z_tf2 = tf.reduce_sum(tf.square(y_tf2))
      dx1 = tf.gradients(z_tf1, x_tf, name='dx1')[0]
      dx2 = tf.gradients(z_tf2, x_tf, name='dx2')[0]
      err = gradient_checker.compute_gradient_error(
          x_tf, x_shape, dx1, x_shape, delta=1e-3, x_init_value=x_np)
      self.assertTrue(err < 1e-3)
      err = gradient_checker.compute_gradient_error(
          x_tf, x_shape, dx2, x_shape, delta=1e-3, x_init_value=x_np)
      self.assertTrue(err < 1e-3)


if __name__ == "__main__":
  tf.test.main()



"""
Basic tests of BFS.

Author: Mengye Ren (mren@cs.toronto.edu)
"""
import numpy as np
import unittest

from tensorflow_forward_ad.cbfs import bfs


class BFSTests(unittest.TestCase):

  def test_bfs(self):
    adj = np.array([[1, 1, 0], [0, 0, 1], [0, 0, 0]], dtype=np.int8)
    self.assertTrue(bfs(adj, 0, 2, None))
    self.assertFalse(bfs(adj, 2, 0, None))

  def test_bfs_cache(self):
    adj = np.zeros([3, 3], dtype=np.int8)
    self.assertTrue(bfs(adj, 0, 2, cache=np.ones([3, 3], dtype=np.int8)))
    self.assertFalse(bfs(adj, 0, 2, cache=np.zeros([3, 3], dtype=np.int8)))

  def test_bfs_performance(self):
    adj = np.random.uniform(0, 1, [1000, 1000])
    adj = (adj > 0.3).astype(np.int8)
    for ii in range(10):
      bfs(adj, 0, 999, None)

  pass


if __name__ == "__main__":
  unittest.main()



import os
import time
import logging

## Plots
import matplotlib
# matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import seaborn as sns

import pandas as pd
import numpy as np

from sklearn.mixture import GaussianMixture
# from scipy.misc import logsumexp

def logsumexp(x,axis):
    return np.log(np.sum(np.exp(x),axis = axis))


def draw_ellipse(position, covariance, num_sig = 1, ax=None, with_center = True, **kwargs):
    """
    Draw an ellipse with a given position and covariance
    Assume diagonal covariance matrix
    """

    ax = ax or plt.gca()

    angle = 0
    width, height = 2 * np.sqrt(covariance)
    x1, x2 = position

    # Draw the Ellipse
    for nsig in range(1, num_sig + 1):
        ax.add_patch(Ellipse(position, nsig * width, nsig * height,
                             angle, **kwargs))


def plot_gmm(alpha, palette, size=1, labels=None, ax=None, mu = None, var = None, gmm = None, hide_legend = False):

    if labels is None:
        labels = np.repeat('Type',mu.shape[0])

    if ax is None:
        _, ax = plt.subplots()

    if gmm is not None:
        mu = gmm.means_
        var = gmm.covariances_
        if mu is None or var is None:
            raise Exception("Either GaussianMixture or a vector of means and variances must be inputted")

    df_plot = pd.concat([pd.DataFrame(mu), pd.DataFrame(labels)], axis = 1)
    dim_labels = ["dim " + str(ii) for ii in range(1,1+mu.shape[-1])]
    df_plot.columns = dim_labels + ["Type"]

    ## Plot
    # Plot contour lines
    for pos, covar, label in zip(mu, var, labels):
        color = palette[label]
        draw_ellipse(pos, covar, ax = ax, color = color, alpha=alpha)

    # Plot the predicted means
    ax = sns.scatterplot(x = dim_labels[0], y = dim_labels[1], s = size,
                         hue = "Type", palette = palette,
                         data = df_plot, linewidth = 0.01)

    if hide_legend:
        ax.get_legend().remove()

    return ax


class GaussianMixtureCustom(GaussianMixture):
    """ Custom Gaussian Mixture, but assuming diagonal covariances"""

    def __init__(self, means, covariances, **kwargs):

        n_components = means.shape[0]
        weights = np.ones(n_components) / n_components
        precisions = 1 / covariances
        precisions_cholesky = 1. / np.sqrt(covariances)

        super().__init__(n_components = n_components, covariance_type = 'diag', **kwargs)

        self.weights_ = weights
        self.means_ = means
        self.covariances_ = covariances
        self.precision_ = precisions
        self.precisions_cholesky_ = precisions_cholesky

    def calculate_negative_likelihood(self, samples):
        nll = -logsumexp(self._estimate_weighted_log_prob(samples),1)
        return nll


    def plot(self, alpha, palette, size=1, labels=None, ax=None, names = None, means = None, covariances = None, max_num_plot = 10, fontsize = 0.1, **kwargs):
        """
        names: vector, (n_samples)
            vector of labels for the samples, which defaults to components
        """

        if means is None:
            means = self.means_
        if covariances is None:
            covariances = self.covariances_

        ax = plot_gmm(mu = means, var = covariances, alpha = alpha, palette = palette, size = size, labels = labels, ax = ax, **kwargs)

        if names is not None:

            ranked_index = self.rank_index(means)

            names_ranked = names[ranked_index]
            means_ranked = means[ranked_index]

            names_plot = names_ranked[:max_num_plot]
            means_plot = means_ranked[:max_num_plot]

            for name, (x,y) in zip(names_plot, means_plot):
                plt.text(x,y,name,fontsize=fontsize)

        return ax

    def rank_index(self, samples):
        """
        Parameters
        ------
        samples: array-like, shape (n_samples, n_features)
        names : vector, length n_samples
        """

        nll = self.calculate_negative_likelihood(samples)
        return nll.argsort()



if __name__ == "__main__":

    nll = -logsumexp(gmm._estimate_weighted_log_prob(means),1)
    ranked_index_high2low = nll.argsort()
    sym_names = np.char.decode(f['sym_names'][:])
    sym_names_cc = sym_names[idx_cell_cycle_noise_filtered]
    sym_names_cc_ranked = sym_names_cc[ranked_index_high2low]

    FUCCIseq_gene_names_sig = ['Cdk1', 'Ube2c', 'Top2a', 'Hist1h4e', 'Hist1h4c']
    np.isin(FUCCIseq_gene_names_sig,sym_names_cc)



import os
import time
import logging

## Plot
import matplotlib
# matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns

import numpy as np
import pandas as pd



def calculate_vmax_vmin(X, cutoff = 0.01):
    """
    X list or arrays of values to be flattened
    """
    assert (cutoff > 0 and cutoff < 0.5), 'cutoff percentage must be between 0 and 0.5'
    flattened_images = np.array([x.reshape(-1) for x in X]).reshape(-1)
    sorted_val = np.sort(flattened_images)
    vmax = sorted_val[int(len(sorted_val) * (1-cutoff))]
    vmin = sorted_val[int(len(sorted_val) * cutoff)]
    return vmax,vmin


def plot_images(X_plots, colnames = None, rownames = None, vmax = None, vmin = None, cutoff = None, center = 0,
                fs_scale = 2, vmaxmin_type = 'individual', cmap = 'RdBu', wspace=0, hspace=0,**kwargs):
    """
    parameters
    ----------
        X_images: col x row x images
        colnames: list of colnames (including the first column)
        rownames: either dict or a list of information for each row
        vmaxmin_type: [all, row, col, individual]  
    """
    #
    plot_args = {}
    #
    ncols = len(colnames)
    nrows = len(rownames)
    #
    if cutoff is not None:
        if vmaxmin_type == 'all':
            vmax, vmin = calculate_vmax_vmin(X_plots, cutoff)
            vmaxs = np.tile(vmax,(nrows,ncols))
            vmins = np.tile(vmin,(nrows,ncols))
        elif vmaxmin_type == 'row':
            list_vals = [[X_plots[icol][irow] for icol in range(ncols-1)] for irow in range(nrows)]
            vmax_vmin = [calculate_vmax_vmin(x,cutoff) for x in list_vals]
            vmax_vmin = np.tile(np.array(vmax_vmin),[ncols,1,1])
            vmax_vmin = np.transpose(vmax_vmin,[2,1,0])
            vmaxs,vmins = vmax_vmin
        elif vmaxmin_type == 'col':
            list_vals = [[X_plots[icol][irow] for irow in range(nrows)] for icol in range(ncols-1)]
            vmax_vmin = [calculate_vmax_vmin(x,cutoff) for x in list_vals]
            vmax_vmin = np.tile(np.array(vmax_vmin),[nrows,1,1])
            vmax_vmin = np.transpose(vmax_vmin,[2,0,1])
            vmaxs,vmins = vmax_vmin
        elif vmaxmin_type == 'individual':
            list_vals = [[calculate_vmax_vmin(X_plots[icol][irow],cutoff) for icol in range(ncols-1)] for irow in range(nrows)]
            vmax_vmin = np.moveaxis(np.array(list_vals),2,0)
            vmaxs,vmins = vmax_vmin
        else:
            raise Exception('Input a valid vmaxmin_type -> [all, row, col, individual]')
    #
    fig, axes = plt.subplots(nrows = nrows, ncols=ncols, sharex=True, sharey=True, figsize = (ncols*fs_scale,nrows*fs_scale))
    #
    for irow in range(nrows):
        logging.info("Row: {}/{}".format(irow+1,nrows))
        for icol in range(ncols):
            logging.info("Col: {}/{}".format(icol+1,ncols))
            ax = axes[irow][icol]
            if irow == 0:
                _ = ax.set_title(colnames[icol])
            if icol == 0:
                if isinstance(rownames, dict):
                    pass
                else:
                    label = rownames[irow]
                    _ = ax.text(7,5,'{}'.format(label), horizontalalignment='center',verticalalignment='center')
                    #
                _ = ax.set_axis_off()
            else:
                X_plot = X_plots[icol-1][irow]
                if X_plot.shape[-1] == 1:
                    ## Set Vmax Vmin
                    if cutoff is not None:
                        plot_args['vmax'] = vmaxs[irow][icol-1]
                        plot_args['vmin'] = vmins[irow][icol-1]
                    ##
                    X_plot = X_plot.sum(-1)
                    g = sns.heatmap(X_plot,ax=ax,cmap=cmap,cbar=False,
                                    xticklabels=False,yticklabels=False, center = center,
                                    **plot_args, **kwargs)
                else:
                    _ = ax.imshow(X_plot)
                    _ = ax.set_axis_off()
            ax.set(adjustable='box', aspect='equal')
    #
    fig.subplots_adjust(wspace=0, hspace=0)
    #
    return fig, axes


def plot_scatter(X_plots, labels_in, dim_labels, colnames = None, rownames = None, vmax = None, vmin = None, cutoff = None, fs_scale = 2,  **kwargs):
    """
    X_images: col x row x images
    colnames: list of colnames (including the first column)
    rownames: either dict or a list of information for each row
    """
    #
    plot_args = {}
    #
    ncols = len(colnames)
    nrows = len(rownames)
    #
    fig, axes = plt.subplots(nrows = nrows, ncols=ncols, sharex=True, sharey=True, figsize = (ncols*fs_scale,nrows*fs_scale))
    #
    for irow in range(nrows):
        logging.info("Row: {}/{}".format(irow+1,nrows))
        for icol in range(ncols):
            logging.info("Col: {}/{}".format(icol+1,ncols))
            ax = axes[irow][icol]
            if irow == 0:
                _ = ax.set_title(colnames[icol])
            if icol == 0:
                if isinstance(rownames, dict):
                    pass
                else:
                    label = rownames[irow]
                    _ = ax.text(7,5,'{}'.format(label), horizontalalignment='center',verticalalignment='center')
                    #
                _ = ax.set_axis_off()
            else:
                X_plot = X_plots[icol-1][irow]
                df_plot = pd.concat([pd.DataFrame(X_plot), pd.DataFrame(labels_in)], axis = 1)
                df_plot.columns = dim_labels + ["Type"]
                # Plot
                sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Type", data = df_plot, **kwargs)
                plt.title(data_type)
                pp.savefig()
                plt.clf()
            ax.set(adjustable='box', aspect='equal')
    #
    return fig, axes


def plot_grid(X_plots, fun_plot, colnames = None, rownames = None, vmax = None, vmin = None, cutoff = None, fs_scale = 2, **kwargs):
    """
    X_images: col x row x images
    colnames: list of colnames (including the first column)
    rownames: either dict or a list of information for each row
    """
    #
    plot_args = {}
    #
    ncols = len(colnames)
    nrows = len(rownames)
    #
    fig, axes = plt.subplots(nrows = nrows, ncols=ncols, sharex=True, sharey=True, figsize = (ncols*fs_scale,nrows*fs_scale))
    #
    for irow in range(nrows):
        logging.info("Row: {}/{}".format(irow+1,nrows))
        for icol in range(ncols):
            logging.info("Col: {}/{}".format(icol+1,ncols))
            ax = axes[irow][icol]
            if irow == 0:
                _ = ax.set_title(colnames[icol])
            if icol == 0:
                if isinstance(rownames, dict):
                    pass
                else:
                    label = rownames[irow]
                    _ = ax.text(7,5,'{}'.format(label), horizontalalignment='center',verticalalignment='center')
                    #
                _ = ax.set_axis_off()
            else:
                X_plot = X_plots[icol-1][irow]
                df_plot = pd.concat([pd.DataFrame(X_plot), pd.DataFrame(labels_in)], axis = 1)
                df_plot.columns = dim_labels + ["Type"]
                # Plot
                sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Type", data = df_plot, **kwargs)
                plt.title(data_type)
                pp.savefig()
                plt.clf()
            ax.set(adjustable='box', aspect='equal')
    #
    return fig, axes



import os
import time
import logging

import numpy as np

from siVAE import FeatureImportance as FI

def residual_analysis(self, y_reconstruct, y_test, min_PVE = 0, take_mean = False, verbose = False, save_PVE = False):
    """
    Calculate PVE and RE of the true and reconstructed numpy array
    Return [RE/PVE, pathway] if y_shape == 3 else [RE/PVE]
    """

    y_residual = y_test - y_reconstruct

    num_nn = y_residual.shape[0]

    output = []

    ## Reconstruction Error calculation
    RE = np.square(y_residual).mean(-2) # MSE per feature
    RE_mean = y_residual.mean(-2) # mean of residual per feature
    RE0 = np.square(y_test).mean(-2) # mean of test value per feature
    logging.info("RE: {}".format(RE))

    if take_mean:
        RE = RE.mean(-1)

    ## PVE calculation
    TV = np.square(y_test.std(axis = -2)) # [.. x sample x gene] -> [.. x gene]
    RV = np.square(y_residual.std(axis = -2)) # [.. x sample x gene] ->[.. x gene]

    if save_PVE:

        PVE = ((TV - RV) / TV)
        if verbose: logging.info("PVE: {}".format(PVE))
        if take_mean:
            PVE = PVE.mean(-1)

        output = (RE, PVE)
    else:
        if TV.shape[0] == 1:
            TV = np.tile(TV,(num_nn,1))
        if RE0.shape[0] == 1:
            RE0 = np.tile(RE0,(num_nn,1))

        output = (RE, RV, TV, RE0, RE_mean)

    return output


def run_FI(input, target, sample_DE, de, sess, method_DE, feed_dict, kwargs_FI = {}, masks = None, mode = 'forward'):
    """
    Run feature attribution
    Parameters
    ----------
    mode
        Run on forward or reverse mode for saliency maps or grad*input
    """

    attributions_dict = {}

    start_time = time.time()

    attributions = FI.feature_attribution(input     = input,
                                          target    = target,
                                          sample    = sample_DE,
                                          de        = de,
                                          sess      = sess,
                                          method_DE = method_DE,
                                          feed_dict = feed_dict,
                                          masks     = masks,
                                          mode      = mode,
                                          **kwargs_FI)

    duration = time.time() - start_time

    attributions_dict['input'] = sample_DE
    attributions_dict['score'] = attributions
    attributions_dict['duration'] = duration
    attributions_dict['methods'] = method_DE

    return attributions_dict




def calculate_decoder_layers(model, sampled_outputs):
    """ Calculate the output of decoder layers """

    decoder_layers_list = []

    for ii in range(len(sampled_outputs)):

        sampled_output = sampled_outputs[ii,]

        feed_dict_full = {model.VAE_sample.X  : sampled_output,
                          model.VAE_sample.X_target : sampled_output,
                          model.VAE_feature.X : model.VAE_dict['feature'].data_handler.X}

        z_mu               = model.VAE_sample.calculate_z_mu(feed_dict_full)
        decoder_layers     = model.VAE_sample.calculate_decoder_layers(feed_dict_full)
        X_mu               = np.array(model.reconstruct_X(feed_dict_full)) # Don't sample X
        decoder_layers     = [z_mu] + decoder_layers
        decoder_layers[-1] = X_mu
        decoder_layers_list.append(decoder_layers)

    return decoder_layers



"""
Implements basic graph algorithms such as BFS.

Author: Mengye Ren (mren@cs.toronto.edu)

Usage:
"""

import numpy as np


def bfs(adj, src, dst, cache=None):
  """BFS search from source to destination. Check whether a path exists, does
  not return the actual path.
  Work on directed acyclic graphs where we assume that there is no path to the
  node itself.

  Args:
    adj: Adjacency matrix.
    src: Source node index, 0-based.
    dst: Destination node index, 0-based.
    cache: 2D matrix, cache[i, j] = 1 indicates path exists between two node
      i and j. cache[i, j] = -1 indicates path does not exists between two node
      i and j. chace[i, j] = 0 indicates unknown.

  Returns:
    found: A path is found between source and destination.
  """
  if src == dst: return False
  num_nodes = adj.shape[0]
  if num_nodes == 0:
    return False
  if src >= num_nodes or dst >= num_nodes:
    raise Exception("Index must be smaller than the number of nodes.")
  if num_nodes == 1:
    return False
  # Whether a node has been visited, if not negative, the parent.
  parent = np.zeros([num_nodes], dtype=np.int64) - 1
  nodes_to_visit = [src]
  found = False
  # BFS loop.
  while len(nodes_to_visit) > 0:
    cur = nodes_to_visit.pop(0)
    if cur == dst:
      found = True
      break
    if cache is not None:
      if cache[cur, dst] == 1:
        found = True
        break
      elif cache[cur, dst] == -1:
        continue

    for jj in range(num_nodes):
      if adj[cur, jj] == 1:
        if parent[jj] == -1:
          nodes_to_visit.append(jj)
          parent[jj] = cur

  if not found:
    # Add the source node to the cache.
    if cache is not None:
      #log.info(("Setting -1", src, dst), verbose=2)
      for ii in range(num_nodes):
        if parent[ii] >= 0:
          cache[ii, dst] = -1
      cache[src, dst] = -1
    return False
  else:
    # Add all the nodes to the cache.
    if cache is not None:
      # Backtrack.
      while cur != src:
        cur = parent[cur]
        cache[cur, dst] = 1
        #log.info(("Setting 1", cur, dst), verbose=2)
      cache[src, dst] = 1
      #log.info(("Setting 1", src, dst), verbose=2)
    return True
